// Copyright (C) 2002 Takeshi Morimoto <morimoto@takopen.cs.uec.ac.jp>
// All rights reserved.
//package sample;

import java.net.*;
import java.util.*;
import yab.agent.*;
import yab.agent.object.*;

public class AmbulanceCenterAgent extends AbstractAmbulanceCenterAgent {
    private ArrayList civilianList = new ArrayList();

    private String REPORT_MESSAGE_CODE = "1";
    private String JOB_ACCEPTED_MESSAGE_CODE = "2";
    private String JOB_ASSIGNED_MESSAGE_CODE = "3";
    private String BUILDINGS_ASSIGNED_MESSAGE_CODE = "4";
    private int ambBuildIndex = 0;
    private boolean allBuildingsAssigned = false;

    private Iterator civIterator = null;
    private Iterator ambIterator;

    public AmbulanceCenterAgent(InetAddress address, int port) {
        super(address, port);

    }

    public void assignBuildings(){

    	int ambCount = 0;
	Logger.writeLog(self().id, "Assigning buildings");

	ambIterator = world.ambulanceTeams.iterator();
	while (ambIterator.hasNext()){
		AmbulanceTeam amb = (AmbulanceTeam)ambIterator.next();
		ambCount++;
	}

	int buildingCount = world.buildings.size();
	int oneBuildingAssignment = buildingCount/ambCount;
	int start = 1+ambBuildIndex*oneBuildingAssignment;
	int end = 1;

	ambIterator = world.ambulanceTeams.iterator();
	for (int i=0;i<ambBuildIndex;i++)
		ambIterator.next();

	Logger.writeLog(self().id, "Ambulance count = "+ambCount);

	for(int i=0;i<4 && ambBuildIndex<ambCount;i++){
		AmbulanceTeam amb = (AmbulanceTeam)ambIterator.next();
		end = start+oneBuildingAssignment;
		tell(BUILDINGS_ASSIGNED_MESSAGE_CODE+" "+amb.id+" "+start+" "+end);
		Logger.writeLog(self().id, "Message "+(i+1)+" "+BUILDINGS_ASSIGNED_MESSAGE_CODE+" "+amb.id+" "+start+" "+end);
		start=end+1;
		ambBuildIndex++;
	}
	if (ambBuildIndex==ambCount)
		allBuildingsAssigned = true;
    }

    public synchronized void assignCivilians(){
    	if (civIterator==null)
		civIterator = civilianList.iterator();
	if (ambIterator==null)
		ambIterator = world.ambulanceTeams.iterator();

	while (civIterator.hasNext() && ambIterator.hasNext()){
		CivilianObject civ = (CivilianObject)civIterator.next();
		AmbulanceTeam amb = (AmbulanceTeam)ambIterator.next();
		tell(JOB_ASSIGNED_MESSAGE_CODE+" "+amb.id+" "+civ.position);
	}

	if (!civIterator.hasNext())
		civIterator = null;
	if(!ambIterator.hasNext())
		ambIterator = null;
    }

    public void act() throws ActionCommandException {
    	Logger.writeLog(self().id, "Time: "+time());
    	if (time()>=4){
		//tell("12345");
		if (!allBuildingsAssigned)
			assignBuildings();
		if (time()%2==0)
			assignCivilians();
	}

    }

    protected synchronized void addToCivilianList(CivilianObject civilian){
    	boolean exists = false;
	for (Iterator iter = civilianList.iterator(); iter.hasNext();){
		CivilianObject civ = (CivilianObject)iter.next();
		if (civ.position.equals(civilian.position))
			exists = true;
	}

	if (!exists)
		civilianList.add(civilian);
    }

    protected void removeCivilian(String civilianPos){
    	Iterator tmpIter = civilianList.iterator();
	while (tmpIter.hasNext()){
		CivilianObject civ = (CivilianObject)tmpIter.next();
		if (civ.position.equals(civilianPos)){
			civilianList.remove(civ);
			return;
		}
	}
    }

    protected void parseAmbMessage(String message){
    	int index = 0;
	int itemIndex = 0;
	int len = message.length();
	String str = "";
	String position = "";
	String hp = "";
	String damage = "";
	String messageType = "";

	while (message.charAt(index)!=' '){
		messageType+=message.charAt(index);
		index++;
	}

	str="";
	index++;
	if (messageType.equals(REPORT_MESSAGE_CODE)){
		while (index<len){
			if (message.charAt(index)!=' ')
				str+=message.charAt(index);
			else{
				if (itemIndex==0) position = str;
				else if (itemIndex==1) hp = str;
				else if (itemIndex==2) damage = str;
				str = "";
				itemIndex++;
				if (itemIndex==3) {
					itemIndex = 0;
					addToCivilianList(new CivilianObject(position, hp, damage, time()));
				}
			}
			index++;
		}
	}
	else if (messageType.equals(JOB_ACCEPTED_MESSAGE_CODE)){
		while(index<len){
			str+=message.charAt(index);
			index++;
		}
		Logger.writeLog(self().id, "Job at target id "+str+" is accepted.");
		removeCivilian(str);
	}
    }

    protected void logCurrentList(){
    	int index = 1;
	for (Iterator iter = civilianList.iterator(); iter.hasNext();){
		CivilianObject civ = (CivilianObject)iter.next();
		Logger.writeLog(self().id, "Civilian"+index+": "+civ.position+" "+civ.hp+" "+civ.damage+" "+civ.time);
		index++;
	}
    }

    protected synchronized void hear(RealObject sender, String message) {
        // do nothing
	if (sender instanceof AmbulanceTeam){
		Logger.writeLog(self().id, sender.id+" "+message);
		parseAmbMessage(message);
		logCurrentList();
	}
    }
}

class CivilianObject{
	public CivilianObject(String _pos, String _hp, String _damage, int _time){
		position = _pos;
		hp = _hp;
		damage = _damage;
		time = _time;
	}

	public String position;
	public String hp;
	public String damage;
	public int time;
}
