// Copyright (C) 2002 Takeshi Morimoto <morimoto@takopen.cs.uec.ac.jp>
// All rights reserved.
//package sample;

import java.net.*;
import java.util.*;
import java.io.*;
import yab.agent.*;
import yab.agent.object.*;

public class AmbulanceTeamAgent extends AbstractAmbulanceTeamAgent {
    private final Set m_unvisitedBuildings;
    private ArrayList humanList = new ArrayList();
    private int EXPLORING = 0;
    private int SEARCHING = 1;
    private int CIVILIAN_FOUND = 2;
    private int CIVILIAN_LOADED = 3;
    private int prev_hp;

    private int MAX_MESS_LENGTH = 80;
    private int MIN_ACC_LIFE = 20;

    protected MotionlessObject previousPos =  null;
    private int state;
    private String targetPosId = "";
    private MotionlessObject targetPos = null;
    private String REPORT_MESSAGE_CODE = "1";
    private String JOB_ACCEPTED_MESSAGE_CODE = "2";
    private String JOB_ASSIGNED_MESSAGE_CODE = "3";
    private String BUILDINGS_ASSIGNED_MESSAGE_CODE = "4";

    Humanoid target = null;
    private int reportCount = 0;

    public AmbulanceTeamAgent(InetAddress address, int port) {
        super(address, port);
        m_unvisitedBuildings = new HashSet(world.buildings);
	state = EXPLORING;
    }

    protected MotionlessObject mlpos() { return self().motionlessPosition(); }

    protected static final Property
        DAMAGE_PRP     = Property.get("Humanoid", "damage"),
        BURIEDNESS_PRP = Property.get("Humanoid", "buriedness"),
        POSITION_PRP   = Property.get("MovingObject", "position"),
        FIERYNESS_PRP  = Property.get("Building", "fieryness");

    protected static final Condition
        IS_NOT_BURNING_CND = FIERYNESS_PRP.lt(3).or(FIERYNESS_PRP.gt(3));

    protected void prepareForAct() {
        super.prepareForAct();
        if (mlpos() instanceof Building)
            m_unvisitedBuildings.remove(mlpos());
    }

    protected void addToCivilianList(Humanoid human){
    	boolean exists = false;
	for (Iterator iter = humanList.iterator(); iter.hasNext();){
		Humanoid hmn = (Humanoid)iter.next();
		if (hmn.id==human.id)
			exists = true;
	}

	if (!exists)
		humanList.add(human);
    }

    /*
    	Buridness'i 0'dan buyuk olan su andaki humanoid'lerin listesi
	geliyor. Bunlari bir listede tutuyoruz. En onemlisini kurtarmaya
	baslayip, digerlerini report ediyoruz.
    */
    protected Humanoid reportBurieds(List brdList) throws ActionCommandException{
    	for (Iterator it = brdList.iterator();it.hasNext();){
		Humanoid hum  =(Humanoid)it.next();
		addToCivilianList(hum);
	}

	String mess = REPORT_MESSAGE_CODE;
	Humanoid target = null;
	double minLife = 10000;

	boolean found = false;
	for (Iterator iter = humanList.iterator(); iter.hasNext() && !found;){
		Humanoid hmn = (Humanoid)iter.next();
		if (hmn instanceof AmbulanceTeam){
			target = hmn;
			found = true;
			continue;
		}
		else if (hmn instanceof FireBrigade){
			target = hmn;
			found = true;
			continue;
		}
		else if (hmn instanceof PoliceForce){
			target = hmn;
			found = true;
			continue;
		}
		if (hmn.damage()!=0){
			double life = hmn.hp()/hmn.damage();
			if (life<minLife){
				minLife = life;
				target = hmn;
			}
		}
	}

	for (Iterator iter = humanList.iterator(); iter.hasNext();){
		Humanoid hmn = (Humanoid)iter.next();

		String tmp = hmn.position().id + " " + hmn.hp() + " " + hmn.damage();
		if (mess.length()+tmp.length()<MAX_MESS_LENGTH){
			if (mess.length()==0)
				mess = tmp;
			else
				mess += " " + tmp;
		}
		if (reportCount==2){
			humanList.remove(iter);
		}
	}

	tell(mess);
	reportCount++;
	if (reportCount==3)
		reportCount = 0;

	/*if (minLife<MIN_ACC_LIFE)
		return target;
	else
		return null;*/
	return target;
	}

    protected void act() throws ActionCommandException {
	Logger.writeLog(self().id,"Time: "+time());
	Logger.writeLog(self().id,"Position: "+mlpos());
	Logger.writeLog(self().id,"My hp: "+self().hp());

	if (prev_hp-self().hp()>100){
		prev_hp = self().hp();
		move(world.refuges);
	}
	else
		prev_hp = self().hp();
	if (state==SEARCHING && mlpos().x()==targetPos.x() && mlpos().y()==targetPos.y())
		state = EXPLORING; //EXploring olacak ama gittigi yerde sivil olacagi
				   //icin target olarak onu atayacak

	if (state==EXPLORING){
		Logger.writeLog(self().id,"Searching for a civilian");

		List hmns = POSITION_PRP.eq(mlpos()).extract(world.humanoids);
		hmns.remove(self());
		Logger.writeLog(self().id, "Civilian count at current pos="+hmns.size());
		if (!hmns.isEmpty()){
			List brd = BURIEDNESS_PRP.gt(0).extract(hmns);
			Logger.writeLog(self().id, "Buried count at current pos="+brd.size());
			if (!brd.isEmpty()){
				target = reportBurieds(brd);
				if (target!=null){
					state=CIVILIAN_FOUND;
					//target = (Humanoid) DAMAGE_PRP.max(brd);
					Logger.writeLog(self().id,"Target civilian is "+target.id);
				}
				else
					searchTheInjured();
			}
			else
				searchTheInjured();
		}
		else
			searchTheInjured();
	}
	else if (state==SEARCHING){
		move(targetPos);
	}
	else if (state==CIVILIAN_FOUND){
		if (target.hp()==0){
		    	Logger.writeLog(self().id,"The civilian "+target.id+" is dead");
		    	state=EXPLORING;
			searchTheInjured();
		}
		else if (target.buriedness()>0){
		 	Logger.writeLog(self().id,"I'm rescuing agent "+target.id+" and buriedness is "+target.buriedness()+" and hp is "+target.hp());
			rescue(target);
		}
		else{
			state = CIVILIAN_LOADED;
			Logger.writeLog(self().id,"Target is not buried now, I'll load it");
			load(target);
		}
	}
	else if (state==CIVILIAN_LOADED){
		if (mlpos() instanceof Refuge){
			Logger.writeLog(self().id,"Arrived to a refuge, civilian is saved");
			state = EXPLORING;
			List hmns = POSITION_PRP.eq(mlpos()).extract(world.humanoids);
			hmns.remove(self());
			for (Iterator it=hmns.iterator();it.hasNext();){
				Humanoid hum = (Humanoid)it.next();
				Logger.writeLog(self().id,"Civilian at this refuge, id="+hum.id+" hp="+hum.hp());
			}
			unload();
		}
		else{
			Logger.writeLog(self().id,"I'm carrying agent to a refuge");
			move(world.refuges);
		}

	}
    }

    protected void carry() throws ActionCommandException {
        if (mlpos() instanceof Refuge)
            unload();
        else
            move(world.refuges);
    }

    protected void rescue(List hmns) throws ActionCommandException{
        List brd = BURIEDNESS_PRP.gt(0).extract(hmns);
        if (!brd.isEmpty())
            rescue((Humanoid) DAMAGE_PRP.max(brd));
    }

    protected void load(List hmns) throws ActionCommandException{
        List inj = DAMAGE_PRP.gt(0).extract(hmns);
        if (!inj.isEmpty())
            load((Humanoid) DAMAGE_PRP.max(inj));
    }

    protected void acceptBuildings(int start, int end){
    	Logger.writeLog(self().id, "Accepting buildings....Start = "+start+" End = "+end);
    	m_unvisitedBuildings.clear();
	int index=1;
	Iterator iter = world.buildings.iterator();
	while (index<start){
		iter.next();
		index++;
	}

	while(index<=end && iter.hasNext()){
		m_unvisitedBuildings.add((Building)iter.next());
		index++;
	}
    }

	protected void searchTheInjured() throws ActionCommandException{
		Logger.writeLog(self().id, "Unvisited building count = "+m_unvisitedBuildings.size());
		List bldgs = IS_NOT_BURNING_CND.extract(m_unvisitedBuildings);
		if (!bldgs.isEmpty()){
			Logger.writeLog(self().id, "There are some buildings that can be moved");
			move(bldgs);
		}
		else
			Logger.writeLog(self().id, "There is no building to move");
	}

	protected void parseMessage(String message){
		int index = 0;
		int itemIndex = 0;
		int len = message.length();
		String str = "";
		String position = "";
		String hp = "";
		String damage = "";
		String messageType = "";

		while (message.charAt(index)!=' '){
			messageType+=message.charAt(index);
			index++;
		}

		str="";
		index++;

		if (messageType.equals(JOB_ASSIGNED_MESSAGE_CODE)){
			while(message.charAt(index)!=' '){
				str+=message.charAt(index);
				index++;
			}
			//Eger mesaj bize geldiyse isle
			if (Integer.parseInt(str)==self().id){
				if (state==EXPLORING){
				 	index++;
					targetPosId = "";
					while (index<len){
						targetPosId+=message.charAt(index);
						index++;
					}
					//targetPos = (MotionlessObject)world.get(Integer.parseInt(targetPosId));
					targetPos = world.get(Integer.parseInt(targetPosId)).motionlessPosition();
					state=SEARCHING;
					Logger.writeLog(self().id, "Job accepted from center. Target id ="+message);
					tell(JOB_ACCEPTED_MESSAGE_CODE+" "+targetPosId);

				}
			}
		}
		else if (messageType.equals(BUILDINGS_ASSIGNED_MESSAGE_CODE)){
			while(message.charAt(index)!=' '){
				str+=message.charAt(index);
				index++;
			}
			//Eger mesaj bize geldiyse isle
			if (Integer.parseInt(str)==self().id){
				index++;
				String start = "";
				while (message.charAt(index)!=' '){
					start+=message.charAt(index);
					index++;
				}
				index++;
				String end = "";
				while (index<len){
					end+=message.charAt(index);
					index++;
				}
				acceptBuildings(Integer.parseInt(start), Integer.parseInt(end));
			}
		}
	}

    protected synchronized void hearTell(RealObject sender, String message) {
        // do nothing
	Logger.writeLog(self().id, "Message received from center="+message);
	if (sender instanceof AmbulanceCenter){
		Logger.writeLog(self().id, "From center: Message received from center="+message);
		parseMessage(message);
	}
    }
}

