import java.lang.Math;
import java.util.*;
import java.net.*;
import java.io.*;

abstract class AgentThd extends Thread implements Definitions
{
	int m_current_state;			//WHICH STATE WE ARE IN
	int m_current_time;
	int m_target_status;			//STATUS FOR CIVILIAN USED BY AMBULANCE
	int m_execute_base=-1;

	byte [] current_msg_body;
	int [] full_message;

	int [] m_state_trace;			//How many times each state is visited
	long [] m_time_trace;			//Record of time steps visited

	int m_agent_type;                       //Agent type used in connecting to kernel, it should stay as integer
	String m_Qvalprefix;
	byte m_agent_index;
	byte m_target_index;
	int m_sent_message_count;
	int m_rcvd_message_count;
	int m_max_send_msg;
	int m_max_rcv_msg;
	//int m_timeout;

	int m_positive_reward;
	int m_id;				//ID of the agent
	MovingObject m_myself;			//Agent itself
	int m_controller_id;
	int m_x;			        //x position of the agent
	int m_y;				//y position of the agent
        int loop_count;
	SimObject we_are_at;			//The object the agent is on
	SimObject random_object;		//The object for the random target
	WorldModel m_world_model;
	LearningComponent m_learner;
	InetAddress m_address;
	IO m_io;
	MessageParser m_msgParser;
	RouteFinder m_router;
	String m_filename;
	Vector m_heard_messages;
	Vector m_sent_messages;
	Vector m_reported_roads;                //The roads that are requested by the RouteFinder to be cleared are stored here, so that same clear is not sent repeatedly
                                                //emptied every DEF_REPORTED_ROAD_CLEAR_TIME
	Vector m_cleared_roads;
        Vector m_reply_buffer;
	boolean current_action_done=false;	//One action per cycle
	
	int last_execute_civilian;		//The id of the civilian
	boolean last_execute_is_job;		//Whether the last civilian is taken as a job(i.e. it has a valid position)
	boolean own_buriedness_sent=false;	//Did the agent send its own buriedness to dispatchers
	Vector m_excluded_targets;		//Visited paths,cleared roads and arrived random targets are stored here, ambulance and police use this in determining new random targets
	                                        //Also roads heard to be clear, cleared roads on the way,jobs that are finished
                                                //Used in considering dispatcher execute messages, considering 'clear please' message.
	Job m_current_job;
	Vector m_job_queue;
	Hashtable injured_civilian_locations;   //Positions arriving with MSG_HELPCIVPOS are stored here by ambulances for further reference in random target search
	int my_training_run_num;
        double m_accept_probability;
        Job job_from_kernel;
	int message_type;                       // Message_type used by HandleMessage : KA_SENSE or KA_HEAR
	boolean canSay;
        boolean canDoubleSay;				
        Job job_to_report;                      //Buffered reply to dispatcher stored here
        MovingObject injured_civ;               //Injured civilian to report stored here
        int sender_id;                          //Sender id of a KA_HEAR message
        HeardInfo heard;                        //Details of the KA_HEAR message
        
        //Variables declared for setMyPosition
        StaticObject pos_head_object;
	StaticObject pos_tail_object;
	int pos_road_length;
	int pos_position_extra;
	double pos_add_x_dist;
        double pos_add_y_dist;
        int pos_head_x;
        int pos_head_y;
        double pos_mult_rate;
        
        //declared for EndOfJob
       	byte saved=NOT_SAVED_YET;
	byte ruined=NOT_RUINED;
        
        //declared for allocateRewards
        double total_reward;
	double police_time_discount;

        //declared for handlePreemption        
        byte [] current_priority_order;
	boolean other_alternatives=false;
	byte priority_of_current_job=-1;
	byte priority_of_environment=-1;
        byte priority_of_dispatcher=-1;
        byte priority_of_own = -1;
	int i;
	Job job_in_queue;
	Job center_job=null;
        byte alt_type;
        
        Building reported_building;
                       
  	AgentThd(int agent_type, String Qvalprefix, String inetaddr, int training_run_num)
	{
	try{
		my_training_run_num=training_run_num;
		m_agent_type=agent_type;
		m_Qvalprefix=Qvalprefix;
		m_address=InetAddress.getByName(inetaddr);
		m_io=new IO(m_address,DEF_SEND_PORT);
		m_world_model=new WorldModel();             //The world information is updated, abstract world parameters are calculated
		m_learner=new LearningComponent();	
		m_msgParser=new MessageParser();
		m_router=new RouteFinder();

		m_sent_message_count=0;
		m_rcvd_message_count=0;
		m_max_send_msg=MAX_MSG_SEND;
		m_max_rcv_msg=MAX_MSG_RCV;
		SetAgentSpecificConstants();
		m_current_time=-1;
		m_heard_messages=new Vector();
		m_sent_messages=new Vector();
		m_reported_roads=new Vector();
		m_job_queue=new Vector();
		//m_target_object=null;
		random_object=null;
		
                m_cleared_roads = new Vector();
		m_excluded_targets=new Vector();
		m_reply_buffer=new Vector();
		injured_civilian_locations=new Hashtable();
		
		m_state_trace=new int [64];	
		m_time_trace=new long[DEF_NUM_TICKS*3+10];
                m_accept_probability = def_q_accept_probability;
                
                loop_count = 0;
	}catch(Exception e){}

	}
	
        abstract void SetAgentSpecificConstants();
	abstract void takeAction();
       	abstract void takeActionRandom();
        abstract void takeActionRandomArrived();
	abstract byte targetSaved();
	abstract byte targetRuined(Job the_job);
	abstract SimObject findMaxSeverity( int m_x, int m_y, int current_time, Job current_target);
        abstract double getStatusOfJob(Job theJob);
        abstract void setStatusOfJob(Job theJob);
        abstract double compareStatuswithOld(Job theJob);
        abstract double compareStatuswithOldUnsaveable(Job theJob);
        
	public void run() 
	{
                        m_accept_probability = def_q_accept_probability; //- ((double)(my_training_run_num / def_discount_interval)) * def_discount_rate;
                
                        m_learner.set_q_filename("Q"+ m_Qvalprefix +".txt");
			m_learner.set_p_filename("P"+ m_Qvalprefix +".txt");
			m_learner.set_Qvisit_filename("QVisit"+ m_Qvalprefix +".txt");
			m_learner.set_Pvisit_filename("PVisit"+ m_Qvalprefix +".txt");
			
                        
                        m_learner.readQvalues();
			m_learner.readPvalues();
                        
			m_id=(int)m_io.connectToKernel(m_agent_type,m_world_model);
			m_controller_id=m_world_model.myController(m_agent_type);
			m_filename="AgentLog"+m_id+".txt";
			m_world_model.setFilename(m_filename);
			m_world_model.setId(m_id);
                        m_learner.set_log_filename(m_filename,m_agent_index);
			m_router.setFilename(m_filename, m_agent_index);
			
			m_myself=(MovingObject)m_world_model.getObjectbyId(m_id,m_agent_index);
			m_myself.m_execute_message=-1;
                        //discoverExecuteMessage();
			SetMyPosition();
			
                        //System.out.println(def_q_accept_probability);
                        //System.out.println(def_p_accept_probability);
                        //System.out.println(def_center_accept_probability);
                        
                
                while(m_current_time<DEF_NUM_TICKS)  //+my_training_run_num)
		{
			handleMessage();
			loop_count ++;	
                        				
		}//End of main loop
                
                
                //GREEDY
		//m_learner.writeQvalues();
                //m_learner.writePvalues();
		calculateOKArea();
		m_learner.writeStateVisitTraces(m_state_trace, m_time_trace, my_training_run_num, loop_count);
		      		
                while (true)
                {               
                      try {wait();} catch (Exception e){}
                }
	}
	
	final void calculateOKArea()
	{
		double total_area=0.0;
		double unburned_area=0.0;
		double this_area=0.0;
		double fieryness;
		Integer temp_index;
		Building temp_building;
		int [] building_indices={WORLD_MODEL_BUILDING_INDEX, WORLD_MODEL_REFUGE_INDEX, WORLD_MODEL_CONTROLLER_INDEX};
		int i;
		
		for (i=0;i<3;i++)
		{	
			Iterator it=m_world_model.the_world[building_indices[i]].keySet().iterator();

			while(it.hasNext())
			{
				temp_index=(Integer)it.next();
				temp_building=(Building)m_world_model.the_world[building_indices[i]].get(temp_index);
				fieryness=temp_building.get_fieryness();
				this_area=temp_building.m_buildingAreaTotal;
				if (fieryness<2)
				{
					unburned_area+=this_area;
				}
				total_area+=this_area;
			}	
		}
		System.out.println("*****************" + unburned_area + "******************");
		System.out.println("*****************" + total_area + "******************");
		m_msgParser.writeMyFile(m_filename,"Unburned Area: " + unburned_area);
		m_msgParser.writeMyFile(m_filename,"Total Area: " + total_area + "\n");
	
	}

	final void handleMessage()
	{	
                full_message=m_io.receiveLongPacket();
		if (full_message==null)
			return;
		message_type=m_msgParser.get_msg_type_long_packet(full_message);
	
		if (message_type==KA_SENSE) 							//world information
		{	
				current_action_done=false;
				resetHeardSaid();
				m_current_time=(int)full_message[MSG_FORMAT_SENSE_TIME];
				//m_time_trace[m_current_time*3]=System.currentTimeMillis(); //Record when we received the sense message...
                                                               
				if ((m_current_time % DEF_REPORTED_ROAD_CLEAR_TIME)==0) 
				{
					m_reported_roads.clear();
					own_buriedness_sent=false;
				}
				//Update world information
				try{
				m_world_model.parseKA_SENSE(full_message, m_current_time);						
				}catch(Exception e){}
					
				SetMyPosition();
                                m_current_state=m_world_model.get_current_state(m_agent_index, m_agent_type);	
				//System.out.println("state" + m_current_state);
                                if (m_current_time>=STEPS_AGENT_FREEZED)
                                {
                                        if (m_current_job!=null )
				        {
			     		      EndOfJob();
				        }
				        removeUnsaveableJobs();              //From job queue

                                        if (!current_action_done && m_myself.get_buriedness()==0)
			                {
			                     handleJobPreemption();	
				             if (m_current_job !=null )	     //We have a goal to satisfy														//And we are not buried
				             {
					        takeAction();
                                                m_time_trace[m_current_time*3 + 1]=System.currentTimeMillis(); //Record when we received the sense message...
				             }       
			                }
                                     
                                        //Keeps track of the visited roads of a police officer
                                        if (m_agent_index==WORLD_MODEL_POLICE_INDEX)
				        {
					     SetVisitedRoads();
				        }
                                }
                               
                                /*
                                if (PRE_LOGGING)
				{
                                     
				 //m_msgParser.writeMyFile(m_filename,"Log AFTER job decision: time: " + m_current_time + " my_id: " + m_id + "my position id:" + m_myself.get_position() + " current state "+ m_current_state + "burning buildings: " + m_world_model.num_burning);
                                 
				 if (m_current_job==null)
				 	;//m_msgParser.writeMyFile(m_filename,"I do not have a target");
				 else
				 	m_msgParser.writeMyFile(m_filename,"Target Type:" +m_current_job.job_type + "Action type(q valueyle belirlenen):"+ m_current_job.q_start_action + "Target id: " +m_current_job.target_id + " previous state: " +m_current_job.q_start_state);
				
                                 if (random_object!=null)
                                        m_msgParser.writeMyFile(m_filename,"But I have a random target:" + random_object.get_id());
                                
                                }
                                */
                                m_state_trace[m_current_state]++;
                                doExtraThings();
				//m_time_trace[m_current_time*3 + 2]=System.currentTimeMillis(); //Record when we received the sense message...
		} 	
				
		else if (message_type==KA_HEAR)
		{
                   if (withinHearLimits()) 
		   {             
			sender_id=full_message[MSG_FORMAT_HEAR_SENDER_ID];
			if ( sender_id==m_controller_id )
			{
				increaseHeardNum();
				if (notHeardBefore(full_message))
				{
					addHeardMessage(full_message);
					heard=m_msgParser.parseKA_HEAR(full_message);
					
                                        //IF THIS IS AN EXECUTE MESSAGE
					//if (heard.detail== m_myself.m_execute_message)
					if (heard.detail == MSG_EXECUTE &&  heard.agent == m_id)
					{	
                                                job_from_kernel=null;        
        					if (!m_excluded_targets.contains(new Integer(heard.subject)))
	               				{
                                                     if (m_current_job != null && m_current_job.target_id == heard.subject)
                                                     {
                                                           //m_msgParser.writeMyFile(m_filename,"KERNELDEN EXECUTE GELDI" + heard.subject + "own job" );
                                                           //Kernel cannot sent the same job twice, thus it is guaranteed that own job is an environment job
                                                           m_current_job.job_type = ACTION_SOURCE_CENTER_ENV;
                                                     }
                                                     else if (SearchTargetinQ(heard.subject))
                                                     {    
                                                           //This debug is done and is ok          
                                                           //m_msgParser.writeMyFile(m_filename,"KERNELDEN EXECUTE GELDI" + heard.subject + "job in our queue" );
                                                     }  
						     else if (m_target_index==WORLD_MODEL_CIVILIAN_INDEX)
						     {
                                                        last_execute_is_job=false;
							last_execute_civilian=heard.subject;
							injured_civ=(MovingObject)m_world_model.the_world[m_target_index].get(new Integer(heard.subject));
							if (injured_civ !=null)
							{       //Originally hps of civilians are 10000 
								injured_civ.m_hp=HELP_REQUESTING_CIVILIAN;  //buraya civilian hp nin initial conditionuna gore biseyler koymak gerek
								if (injured_civ.m_position>0)
								{
						          		last_execute_is_job=true;
                                                                        job_from_kernel=new Job((int)heard.subject,ACTION_SOURCE_CENTER);
                                                                        injured_civ.m_assignedTo = 1; //In order not to report own target as an injured civilian
							         	job_from_kernel.target_object=injured_civ;
							 		setStatusOfJob(job_from_kernel);
                                                                        putToJobQueue(job_from_kernel);
                                                                        job_from_kernel=null;
								}
							}	
			                              }
						      else if (m_target_index==WORLD_MODEL_BUILDING_INDEX)
						      { 
                                                            job_from_kernel=new Job((int)heard.subject,ACTION_SOURCE_CENTER);
						            job_from_kernel.target_object=(SimObject)m_world_model.the_world[m_target_index].get(new Integer(heard.subject));
							        
                                                            if (m_world_model.extinguished(job_from_kernel.target_object))
                                                            {
                                                                 /*This case is tested and is successful   
                                                                 m_msgParser.writeMyFile(m_filename,"Dispatcher job already extinguished"+ heard.subject);
                                                                 */
                                                                 handleCenterReply(job_from_kernel, MSG_NEUTRAL_POS);
                                                                 job_from_kernel=null;                
                                                            }
                                                            else if (m_world_model.burntDown(job_from_kernel.target_object))
                                                            {
                                                                 /*This case is tested and is successful   
                                                                 m_msgParser.writeMyFile(m_filename,"Dispatcher job already burnt down"+ heard.subject);
                                                                 */
                                                                 handleCenterReply(job_from_kernel, MSG_NEUTRAL_NEG);
                                                                 job_from_kernel=null;
                                                            }
                                                            else
                                                            {
							         setStatusOfJob(job_from_kernel);
							         //m_msgParser.writeMyFile(m_filename,"KERNEL JOB PUT IN Q " + heard.subject);
							         putToJobQueue(job_from_kernel);
                                                                 job_from_kernel=null;
                                                            }
							 }	
                                                         else if (m_target_index==WORLD_MODEL_ROAD_INDEX)
                                                         {
                                                                 job_from_kernel=new Job((int)heard.subject,ACTION_SOURCE_CENTER);
							         job_from_kernel.target_object=(SimObject)m_world_model.the_world[m_target_index].get(new Integer(heard.subject));
							       
                                                                 if (!((Road)job_from_kernel.target_object).confirmed)
                                                                 {
                                                                        m_world_model.updateWorld(heard.subject, m_target_index, MAX_SEVERE_ROAD);
                                                                        //m_msgParser.writeMyFile(m_filename,"POL" + m_id + " DISP JOB WILL BE IN Q"+ heard.subject);
                                                                        //System.out.println("POL" + m_id + " DISP JOB WILL BE IN Q"+ heard.subject);
                                                                 }
							         else if (m_world_model.clear(job_from_kernel.target_object))
                                                                 {                                                                        
                                                                        //m_msgParser.writeMyFile(m_filename,"Dispatcher job already clear"+ heard.subject);
                                                                        handleCenterReply(job_from_kernel, MSG_NEUTRAL_POS);
                                                                 }
                                                                        
                                                                 if (!m_world_model.clear(job_from_kernel.target_object))
                                                                 {
                                                                        setStatusOfJob(job_from_kernel);
							                //m_msgParser.writeMyFile(m_filename,"POL" + m_id + " DISP JOB WILL BE IN Q 2"+ heard.subject);
                                                                        //System.out.println("POL" + m_id + " DISP JOB WILL BE IN Q 2"+ heard.subject);
							                putToJobQueue(job_from_kernel);
                                                                 }
                                                                 job_from_kernel=null;
                                                         }
						     }
                                                     else  //finished job vector contains this job
                                                     {
                                                                //This debug is done and is ok.
                                                                //System.out.println("Already finished job sent..." + heard.subject);
                                                               
                                                                job_from_kernel=new Job((int)heard.subject,ACTION_SOURCE_CENTER);
							        job_from_kernel.target_object=(SimObject)m_world_model.the_world[m_target_index].get(new Integer(heard.subject));
							       
                                                                if (targetRuined(job_from_kernel) == RUINED)
                                                                        handleCenterReply(job_from_kernel, MSG_NEUTRAL_NEG);
                                                                else
                                                                {
                                                                        handleCenterReply(job_from_kernel, MSG_NEUTRAL_POS);
                                                                }
                                                                job_from_kernel = null;
                                                     }
						}
						//else if (m_target_index==WORLD_MODEL_CIVILIAN_INDEX && heard.detail==(MSG_EXPOS + (m_myself.m_execute_message-MSG_AMBULANCE_EXECUTE)))
						else if (m_target_index==WORLD_MODEL_CIVILIAN_INDEX && heard.detail==MSG_EXPOS && heard.agent== m_id)
						{
							if (last_execute_civilian>0)
							{
								injured_civ=(MovingObject)m_world_model.the_world[m_target_index].get(new Integer(last_execute_civilian));
								if (injured_civ !=null)
								{
									injured_civ.m_position=heard.subject;
									if (last_execute_is_job==false)
									{
										job_from_kernel=new Job(injured_civ.m_id,ACTION_SOURCE_CENTER);
										job_from_kernel.target_object=injured_civ;
										injured_civ.m_hp=HELP_REQUESTING_CIVILIAN;
										setStatusOfJob(job_from_kernel);
										putToJobQueue(job_from_kernel);
										last_execute_is_job=true;
									}	
								}				
							}
						}
						else if (heard.detail==  MSG_CLEAR)
						{
                                                        //We shall not update a cleared road as blocked
							if (!m_excluded_targets.contains(new Integer(heard.subject)))
							{
								//Just updating our world
								m_world_model.updateWorld(heard.subject,WORLD_MODEL_ROAD_INDEX,MAX_SEVERE_ROAD);
							}
						}
						
						else if (heard.detail==MSG_ROAD_CLEARED)
						{
							m_world_model.updateWorld(heard.subject,WORLD_MODEL_ROAD_INDEX,0);
							m_excluded_targets.add(new Integer(heard.subject));
						}
						
						else if (heard.detail==MSG_HELPCIVPOS)
						{
							if (m_target_index==WORLD_MODEL_CIVILIAN_INDEX)
							{
								if (!injured_civilian_locations.containsKey(new Integer(heard.subject)))
								{
									injured_civilian_locations.put(new Integer(heard.subject),m_world_model.getObjectbyId(heard.subject));
									//System.out.println("ambulance is storing a heard civilian location");	
								}
							}
						}	
                                                
                                                else if (heard.detail== MSG_FIERY_BUILDING)
                                                {
                                                        reported_building =(Building) m_world_model.the_world[WORLD_MODEL_BUILDING_INDEX].get(new Integer(heard.subject));
                                                        if (reported_building.m_fire_reported==0)
                                                        {
                                                                if (reported_building.m_fieryness ==0) reported_building.m_fieryness = 2;
                                                                m_world_model.buildings_tobe_reported.remove(reported_building);
                                                                reported_building.m_fire_reported = 1;
                                                        }
                                                        //System.out.println("Agent " + m_id + " heard building " + reported_building.m_id);
                                                }
		  	  }//end of msg received before control
		      }//end of sender id control
                      
                  }//end of received msg count control	
                  
 		}//end of msg_type if
		
	}
        
        final void doExtraThings()
        {
                
                canSay = withinSayLimits();
                canDoubleSay = withinDoubleSayLimits();
                
               	//REPORT REPLY TO DISPATCHER
		if (canSay && m_reply_buffer.size()>0)
		{
			job_to_report=(Job)m_reply_buffer.remove(0);
			//JOB type contains success/fail/neutral
                        //m_msgParser.writeMyFile(m_filename,"Dispatcher reply sent from buffer for " + job_to_report.target_id);
			m_io.sendAK_TELL(m_id, job_to_report.job_type, job_to_report.target_id);   
			increaseSaidNum();	
		}
                //REPORT OWN INJURY
		else if ( canDoubleSay && m_myself.get_buriedness()>0 && own_buriedness_sent==false)
		{		
			m_io.sendAK_TELL(m_id, MSG_HELPCIVPENDINGPOS, m_id);
			increaseSaidNum();
			m_io.sendAK_TELL(m_id, MSG_HELPCIVPOS, we_are_at.get_id());
			increaseSaidNum();
			own_buriedness_sent=true;				
	       }
               //REPORT OTHERS INJURY
	       else if (canDoubleSay && m_world_model.num_injured>0)
	       {
			injured_civ=(MovingObject)m_world_model.findACivilianToReport();
			if (injured_civ != null)
			{
				m_io.sendAK_TELL(m_id, MSG_HELPCIVPENDINGPOS, injured_civ.get_id());
				increaseSaidNum();
				m_io.sendAK_TELL(m_id, MSG_HELPCIVPOS, injured_civ.get_position());
				increaseSaidNum();
			}
		
	       }
               //REPORT FIERY BUILDING
               else if (canSay && m_world_model.buildings_tobe_reported.size()>0)                
               {
                        reported_building = (Building)m_world_model.buildings_tobe_reported.remove(0);
                        m_io.sendAK_TELL(m_id,MSG_FIERY_BUILDING ,reported_building.m_id);
                        increaseSaidNum();	
                        reported_building.m_fire_reported = 1;
                        //System.out.println ("Agent " + m_id + " reported building " + reported_building.m_id);
               }
               //REPORT CLEARED ROAD
               else if (canSay && m_cleared_roads.size()>0)
               {
                        if (PRE_LOGGING)m_msgParser.writeMyFile(m_filename,"REPORTING cleared road ");
			m_io.sendAK_TELL(m_id,MSG_ROAD_CLEARED ,((Integer)m_cleared_roads.remove(0)).intValue());
			increaseSaidNum();		       
               }
	       //REPORT YOUR POSITION
	       else if(canSay)
	       {
	       		if (PRE_LOGGING)m_msgParser.writeMyFile(m_filename,"yes we are updating position");
			m_io.sendAK_TELL(m_id, MSG_POSITION, m_myself.get_position());   
			increaseSaidNum();
	       }			
	       
			
        }
        
        final boolean notHeardBefore(int [] hear_message)
	{
		//return (!m_heard_messages.contains(hear_message));
                
                String str;
                int i;
                str = "";
                for (i=0; i< hear_message.length; i++)
                {
                        str = str + "" + hear_message [i];
                }
                
                if (!m_heard_messages.contains(str))
                {
                        return true;
                }
                else
                {
                        return false;
                }
	}
	
	final void addHeardMessage(int [] hear_message)
	{
                String str;
                int i;
                str = "";
                for (i=0; i< hear_message.length; i++)
                {
                        str = str + "" + hear_message [i];
                }
                
		m_heard_messages.add(str);
	}

	final void SetMyPosition()
	{
	       	we_are_at = (SimObject)m_world_model.getObjectbyId( m_myself.get_position());
		
		if (we_are_at instanceof StaticObject && !(we_are_at instanceof Road))
		{
			m_x=((StaticObject)we_are_at).get_x();
			m_y=((StaticObject)we_are_at).get_y();		
		}
		else if (we_are_at instanceof Road)
		{	
			pos_position_extra=m_myself.get_positionExtra();
			pos_road_length=((Road)we_are_at).get_length();
                        
			pos_head_object=(StaticObject)m_world_model.getObjectbyId(((Road)we_are_at).get_head());
			pos_tail_object=(StaticObject)m_world_model.getObjectbyId(((Road)we_are_at).get_tail());
			
                        pos_head_x = pos_head_object.get_x();
                        pos_head_y = pos_head_object.get_y();
                        pos_mult_rate = (double)pos_position_extra/(double)pos_road_length;
                        
			pos_add_x_dist=((double)(pos_tail_object.get_x()-pos_head_x))*pos_mult_rate;
			pos_add_y_dist=((double)(pos_tail_object.get_y()-pos_head_y))*pos_mult_rate;
			
			if (IS_LOGGING) m_msgParser.writeMyFile(m_filename,"position_extra" + pos_position_extra + "road_length" + pos_road_length +"add_x_dist" + pos_add_x_dist +"add_y_dist" + pos_add_y_dist );
			
			m_x=pos_head_x+(int)pos_add_x_dist;
			m_y=pos_head_y+(int)pos_add_y_dist;
		}
		else if (we_are_at instanceof MovingObject)//if we are on a moving object:AMBULANCE , we cannot do any work so we skip this round
		{
			current_action_done=true;
		}

	}
	
	
	/**  Adds the roads a police agent has passed to m_excluded_targets so that police agent does not pass them 
	  *  when moving around randomly. To increase map coverage
	  */
        /// |0| | | | | | | |size-1|
        //   |                  |
        //   |                  |
        //  target              position
        
	final void SetVisitedRoads()
	{
		int i;
		int routeNodeId;
		
		for (i=m_router.m_position_history.size()-1;i>0; i--)		
			{
				//System.out.println("Police gittigi yollari store ediyor");
				routeNodeId=((RouteNode)m_router.m_position_history.elementAt(i)).Id;
				if (routeNodeId!=we_are_at.get_id())
				{
					m_excluded_targets.add(new Integer(routeNodeId));
					//GECTIGIMIZ YERLERI CLEAR KABUL EDICEZ....
					//TERS YONE BLOCK OLSA BILE..
					//CENTER JOBU ONLARI TEKRAR BLOCKED YAPIYO ZATEN
					//m_world_model.updateWorld(routeNodeId,WORLD_MODEL_ROAD_INDEX,0);		
                                        //10/12/2002 the above lines are invalidated. One-way passable set edilebilir...
				}
				else
				{	
					//If we have moved we add our current position
					if (i!=m_router.m_position_history.size()-1)
					{
						m_excluded_targets.add(new Integer(routeNodeId));
					}
                                        //We do not go beyond our own position
					break;
				}
			}
		
	}

	//Finds a path to ObjectId,removelast indicates whether target itself is removed from path
	final boolean moveTo(int ObjectId,  boolean removelast)
	{
          	if (!current_action_done && m_current_time>=STEPS_AGENT_FREEZED)
	  	{
	  
	  
			RouteResult routeReply;

			routeReply=m_router.FindRouteBetween(we_are_at, ObjectId, m_world_model, removelast,m_agent_index==WORLD_MODEL_POLICE_INDEX, m_myself.get_positionExtra());
                        m_router.reset_next_cycle=false;
			if (routeReply.route!= null)
			{
                                if (m_agent_index==WORLD_MODEL_POLICE_INDEX && routeReply.route.length>=2)
                                 {
                                        //If our route includes the blocked neighbor, we shall first send clear for that road
                                        //This control was done afterwards the route was sent by the police officer
                                        //In that way we lost a guaranteed 1 cycle
                                        if (routeReply.blockedNeighbor == routeReply.route[0] || routeReply.blockedNeighbor == routeReply.route[1]) 
                                        {
                                                m_io.send_two_int_msg(AK_CLEAR,m_id, routeReply.blockedNeighbor);   
                                                current_action_done = true;
                                                m_router.reset_next_cycle=true;
                                        }
                                        else
                                        {
                                                m_router.reset_next_cycle=false;
                                        }
                                }
                                
                                if (!current_action_done)
                                {
        				m_io.sendAK_MOVE(m_id,routeReply.route);
                                        current_action_done = true;       
                                }        
			}
	
			//We can have a route and also a CLEAR_ROUTE command
			if (routeReply.routerResponse==CLEAR_ROUTE)
			{
				if (m_agent_index!=WORLD_MODEL_POLICE_INDEX)
				{
					if (withinSayLimits()&& !(m_reported_roads.contains(new Integer (routeReply.blocked))) )
					{
                                                //m_msgParser.writeMyFile(m_filename,"FF" + m_id + " asking road to be cleared"+ routeReply.blocked);
                                                //System.out.println("FF" + m_id + " asking road to be cleared"+ routeReply.blocked);
						m_io.sendAK_TELL(m_id, MSG_CLEAR, routeReply.blocked);   
						increaseSaidNum();
						m_reported_roads.add(new Integer (routeReply.blocked) );
					}
					random_object=null; //If you are blocked while searching a random target, drop it
				}
				else
				{
                                        if (!current_action_done)
                                        {
                                                m_io.send_two_int_msg(AK_CLEAR,m_id, routeReply.blocked);
                                                current_action_done = true;
                                        }
                                        // System.out.println( " " + m_id + "Problem problem. Police officer reports blocked road...");
					//A POLICE DOES NOT REPORT A BLOCKAGE.
					//THE BLOCKAGES THAT CAN BE HANDLED BY A POLICE ARE HANDLED BEFOREHAND.
					//THUS A POLICE OFFICER SHOULD NOT SEND A CLEAR COMMAND ALSO.
				}
			}
			else if (routeReply.routerResponse==DROP_JOB)
			{
				m_world_model.ingoredJobsofPolice.add(new Integer(routeReply.blocked));
				if (m_current_job != null)
				{
					handleCenterReply(m_current_job, MSG_NEUTRAL_NEG);
					resetJobIndicators(m_current_job.target_id);
				}
				else
				{
					random_object=null;
				}
			}
                        
                        /*This condition is checked and found that route finder returns NULL_ROUTE2 : No route found and we are at level 1:  cannot go e
			else if (routeReply.route == null)
			{
                                m_msgParser.writeMyFile(m_filename,"moveTo: Route is null but no blocked road? ROUTER RESPOSNE IS: " + routeReply.routerResponse);
                        }
		        */
                        
	  		//IF THE AGENT NEEDS TO MOVE TO DO ITS JOB THEN IT CANNOT DO ANY OTHER ACTION 
			//EVEN IF IT IS NOT POSSIBLE FOR IT TO MOVE
	   		//WE WILL SET current_action_done TO TRUE TO PRESERVE RESOURCES OF COMPUTER 27/4/2002
	    		current_action_done=true;
                        
			return true;
	    }//end of action done check
	    
	    return false;
	}

	//Finds the angle between two points, used in extinguish function
	final int direction(int startx,int starty, int endx, int endy)
	{
		int xdist=endx-startx;
		int ydist=endy-starty;

		double angle=Math.atan2(-xdist,ydist); //origin is at positive y direction
		if (angle<0) angle+=2*Math.PI;
		return(int) (angle*360*60*60/(2*Math.PI));
	}
	
	final void EndOfJob()
	{	
		saved=targetSaved();
		ruined=targetRuined(m_current_job);
                	
			if(saved== SAVED_BY_ME || saved== SAVED_BY_OTHER || ruined==RUINED )
			{
                                if (m_agent_index==WORLD_MODEL_FIREFIGHTER_INDEX)
		                {
		                      m_positive_reward=((Building)(m_current_job.target_object)).m_buildingAreaTotal /1000;
                                      //System.out.println("My reward is " + m_positive_reward);
		                } 
                                
				if (PRE_LOGGING) m_msgParser.writeMyFile(m_filename,"Elimizdeki is bitti");
				m_excluded_targets.add(new Integer(m_current_job.target_id));
				
				if (saved== SAVED_BY_ME)
				{
                                        //m_msgParser.writeMyFile(m_filename,"JOB SAVED" + m_current_job.target_id);
					allocateRewards(m_current_job, m_positive_reward, INTERMEDIATE_REWARD, MSG_SUCCESS);
						
					if (m_agent_index==WORLD_MODEL_POLICE_INDEX)
					{
                                                m_cleared_roads.add(new Integer(m_current_job.target_id));
					}
					
				}
                                //The problem of unfinished jobs at the end of 300 cycles is solved by increasing the training runs to 500 cycles
				//else if(ruined==RUINED || (m_current_time==DEF_NUM_TICKS && saved==NOT_SAVED_YET && m_current_job.cycles_spent_on_job>=m_timeout))
                                //CHANGE FOR NEWEST: NEGATIVE REWARD IS NOT GIVEN. NEGATIVE IS NO REWARD
                                
                                else if(ruined==RUINED)
                                {		
                                        //m_msgParser.writeMyFile(m_filename,"JOB RUINED" + m_current_job.target_id);        
					allocateRewards(m_current_job, -m_positive_reward, -INTERMEDIATE_REWARD, MSG_FAILURE);				
				}
				else //if saved by other
                                
				{	
				
					allocateRewards(m_current_job, 0, INTERMEDIATE_REWARD, MSG_NEUTRAL_POS);
					
				}//end of good & bad if	
						
				resetJobIndicators(m_current_job.target_id);	
			}//end of job finished if
			
	}//end of endofjob
	
	
	final void allocateRewards(Job current_job, int actual_reward, int intermediate_reward, int center_reply)
	{
		total_reward=0.0;
		police_time_discount=0.0;
		
		if (m_agent_index==WORLD_MODEL_POLICE_INDEX)
		{
			police_time_discount=current_job.cycles_spent_on_job * POLICE_DELAY_DISCOUNT_RATE;
		}
                				
		if ((current_job.job_type & ACTION_SOURCE_CENTER )>0)
		{
			handleCenterReply(current_job, center_reply);
                        //if (current_job.job_type == ACTION_SOURCE_CENTER_ENV)
                                //System.out.println("Dispatcher response sent for job " + current_job.target_id);
		}
                
		if (actual_reward>0)
                {
		      if (current_job.q_start_action>=0)
		      {
                        //m_msgParser.writeMyFile(m_filename,"Q val update. Target " +  current_job.target_id + " decision by " + current_job.q_start_action);      
			m_learner.updateQvalues(current_job.q_start_state,current_job.q_start_action, m_current_state, actual_reward-police_time_discount, m_current_time);	
		      }
					
		      if (current_job.preempt_hist_index>=0)
		      {
                        total_reward=rewardFromQueue();	
                        //if (m_agent_index==WORLD_MODEL_POLICE_INDEX)
                        //        System.out.println("" + m_id  + "- actual reward: " + actual_reward + " from queue: " + total_reward + "accumulated: " + current_job.accumulated_reward + "discount: " + police_time_discount);

			total_reward+=current_job.accumulated_reward; 
			total_reward+=actual_reward;   
			total_reward-=police_time_discount;
                      
			DegradingReward(current_job,total_reward);								
		      }
                }
	}
	
        final void allocateRewardsQJob(Job current_job, int actual_reward)
	{
		total_reward=0.0;
		police_time_discount=0.0;
		
		if (m_agent_index==WORLD_MODEL_POLICE_INDEX)
		{
			police_time_discount=current_job.cycles_spent_on_job * POLICE_DELAY_DISCOUNT_RATE;
		
		}
		
                if (actual_reward>0)
                {
		      if (current_job.q_start_action>=0)
		      {
                        //m_msgParser.writeMyFile(m_filename,"Q val update queue. Target " +  current_job.target_id + " decision by " + current_job.q_start_action);            
			m_learner.updateQvalues(current_job.q_start_state,current_job.q_start_action, m_current_state, actual_reward-police_time_discount, m_current_time);	
		      }
					
		      if (current_job.preempt_hist_index>=0)
       		       {
			//Different from the previous function we do not include the Reward from queue part here
			total_reward+=current_job.accumulated_reward;
			total_reward+=actual_reward;   
			total_reward-=police_time_discount;

			DegradingReward(current_job,total_reward);
		      }
                }
	}
	
        //This function is just for debugging
	final void WritePHist(Job current_job)
	{
		int i,j;
		
		PreemptionHistory hist_item;
		//if (m_id==2327)
		//	System.out.println("size:" + current_job.preempt_hist_index + " id:" + current_job.target_id);
		//Starting from more recent, going back
		for (i=current_job.preempt_hist_index;i>=0;i--)
		{
			hist_item=current_job.preempt_hist[i]; 
			if (hist_item.preemption_alternative!=ALT_NO_ALT)
			{
				if (hist_item.preemption_direction==TAKE_OUT)
				{
				//if (m_id==2327)
				//		System.out.println("TAKE_OUT updating reward" + hist_item.job_type);
			    }
				else
				{
					//if (m_id==2327) System.out.println("PUT_IN begin our type: " + hist_item.job_type);
					//This time the ones with higher priority get the reward (positive or negative)
					for(j=0;j<def_action_source_num;j++)
					{
						if (hist_item.preemption_order[j]>hist_item.preemption_order[hist_item.job_type])
						{
							//if (m_id==2327)
							//	System.out.println("	PUT_IN updating reward" + j);
						}
					}
				}
			}
		}
	}
	
	final void DegradingReward(Job current_job, double total_reward)
	{
		int i,j;
		double reward_multiplier=1.0;
		PreemptionHistory hist_item;
		
		//Starting from more recent, going back
		for (i=current_job.preempt_hist_index;i>=0;i--)
		{
			hist_item=current_job.preempt_hist[i]; 
			if (hist_item.preemption_alternative!=ALT_NO_ALT)
			{
				if (hist_item.preemption_direction==TAKE_OUT)
				{
                                        //m_msgParser.writeMyFile(m_filename,"Rewarding take out for " + hist_item.job_type + " state " + hist_item.preemption_state );
					m_learner.updatePvalues(hist_item.preemption_state,hist_item.job_type, m_current_state, reward_multiplier*total_reward,m_current_time, hist_item.preemption_alternative, reward_multiplier);
			        }
				else
				{
                                        if (hist_item.job_type == ACTION_SOURCE_CENTER_ENV)
                                        {
                                                //m_msgParser.writeMyFile(m_filename,"Rewarding put back for ACTION_SOURCE_OWN state " + hist_item.preemption_state );
                                                m_learner.updatePvalues(hist_item.preemption_state,ACTION_SOURCE_OWN, m_current_state, reward_multiplier*total_reward,m_current_time, hist_item.preemption_alternative, reward_multiplier);
                                        }
                                        else if (hist_item.job_type == ACTION_SOURCE_CENTER_OWN)
                                        {
                                                //m_msgParser.writeMyFile(m_filename,"Rewarding put back for ACTION_SOURCE_ENVIRONMENT state " + hist_item.preemption_state );
                                                m_learner.updatePvalues(hist_item.preemption_state,ACTION_SOURCE_ENVIRONMENT, m_current_state, reward_multiplier*total_reward,m_current_time, hist_item.preemption_alternative, reward_multiplier);
                                        }
                                        else
                                        {
					   //This time the ones with higher priority get the reward (positive or negative)
					   for(j=0;j<def_action_source_num;j++)
					   {
                                               
						if (hist_item.preemption_order[j]>hist_item.preemption_order[hist_item.job_type])
						{
                                                        //m_msgParser.writeMyFile(m_filename,"Rewarding put back for " + j + " state " + hist_item.preemption_state );
							m_learner.updatePvalues(hist_item.preemption_state,j, m_current_state, reward_multiplier*total_reward,m_current_time, hist_item.preemption_alternative, reward_multiplier);
						}
					   }
                                        }
				}
				reward_multiplier/=2;
			}
		}
	}
	
	final void resetJobIndicators(int this_target)
	{
		int i;
		Job tempjob;
		
		for (i=0;i<m_job_queue.size();i++)
		{
			tempjob=(Job)m_job_queue.elementAt(i);
			if (tempjob.target_id==this_target)
			{
				removeFromJobQueue(tempjob);
				i--; //Since the indexes have shifted...
			}
		}
		
		m_current_job=null;
		
		if (m_target_index==WORLD_MODEL_CIVILIAN_INDEX)
		{
			m_target_status=AMBULANCE_TARGET_STATUS_NOT_FOUND;
		}
	}
	
	final boolean withinHearLimits()
	{
		return (m_rcvd_message_count<m_max_rcv_msg);
	}
	
	final boolean withinSayLimits()
	{
		return (m_sent_message_count<m_max_send_msg);
	}
	
	final boolean withinDoubleSayLimits()
	{
		return (m_sent_message_count<m_max_send_msg-1);
	}
	
	final void increaseHeardNum()
	{
		m_rcvd_message_count++;
	}
	
	final void increaseSaidNum()
	{
		m_sent_message_count++;
	}
	
	final void resetHeardSaid()
	{
		m_rcvd_message_count=0;
		m_sent_message_count=0;
	}
	               
        final int numOfSaveableTargets()
	{   
                return m_world_model.targets_tobe_saved.size();
	}
	
	final void removeFromJobQueue(Job theJob)
	{
		m_job_queue.remove(theJob);
	}
	
	final void putToJobQueue(Job theJob)
	{
		m_job_queue.add(theJob);
	}
	
	
	final Job preemptJob(Job newJob, Job oldJob)
	{
		putToJobQueue(oldJob);
		removeFromJobQueue(newJob);
                return newJob;
	}
	
	final byte getPriorityofJob(Job theJob,byte [] current_priority_order)
	{
		return current_priority_order[theJob.job_type];
	
	}
	
	final byte getPriorityofJob(int action_source,byte [] current_priority_order)
	{
		return current_priority_order[action_source];
	
	}
	
        //Called after KA_SENSE is received, before new job is determined
	void removeUnsaveableJobs()
	{
		Job job_in_queue;
		int i;
		double rewardofjobinQ;
		double accumulated_reward=0;
			
		for (i=0;i<m_job_queue.size();i++)
		{
			job_in_queue=(Job)m_job_queue.elementAt(i);
			if (!(m_world_model.saveable(job_in_queue.target_object)))
			{
                                //Center replies are sent from here
				rewardofjobinQ=compareStatuswithOldUnsaveable(job_in_queue);
                                accumulated_reward+=rewardofjobinQ;
                                
                                //CHANGE FOR NEWEST:NEGATIVE REWARD IS NOT GIVEN
                                
                                //If this job was ruined because of being in the queue, we shall allocate necessary rewards
                                if (rewardofjobinQ<0)
                                {
                                        if (m_agent_index==WORLD_MODEL_FIREFIGHTER_INDEX)
		                        {
		                              m_positive_reward=((Building)(job_in_queue.target_object)).m_buildingAreaTotal /1000;
		                        }
                                        allocateRewardsQJob(job_in_queue, -m_positive_reward);
			        }
                                
                                //If the job was removed in compareStatuswithOldUnsaveable, we go back one step, since upon removal indices shift back
				if (i!=m_job_queue.indexOf(job_in_queue))
				{
					i--;
				}
			}
		}
	
		if (m_current_job!=null)
		{
			m_current_job.accumulated_reward+=accumulated_reward;
		}
	}
	
	final void handleCenterReply(Job theJob, int theResult)
	{
		if ((theJob.job_type & ACTION_SOURCE_CENTER)>0)
		{
			if (withinSayLimits())
			{
                                //m_msgParser.writeMyFile(m_filename,"Dispatcher reply sent for " + theJob.target_id);
				m_io.sendAK_TELL(m_id,theResult ,theJob.target_id);
				increaseSaidNum();	
			}
			else
			{	
		              	theJob.job_type=(byte)theResult;
				m_reply_buffer.add(theJob);
			}	
		}
	}
	
	int rewardFromQueue()
	{
		Job job_in_queue;
		int i;
		double rewardofjobinQ;
		int rewardofCurrent=0;
	
            
		for (i=0;i<m_job_queue.size();i++)
		{
			job_in_queue=(Job)m_job_queue.elementAt(i);
			rewardofjobinQ=compareStatuswithOld(job_in_queue);
			rewardofCurrent+=rewardofjobinQ;
			//The job was removed, so we shall take the index back
			if (i!=m_job_queue.indexOf(job_in_queue))
			{
				i--;
			}
			/*do not make this update 29/10/2002
			if (job_in_queue.preempt_hist_index>=0)  //Kernel joblarinki -1 olabiliyor
			{
				//m_learner.updatePvalueIntermediate(job_in_queue.preempt_hist[job_in_queue.preempt_hist_index].preemption_state,job_in_queue.job_type,rewardofjobinQ,job_in_queue.preempt_hist[job_in_queue.preempt_hist_index].preemption_alternative);
				DegradingReward(job_in_queue,rewardofjobinQ);
			}
                        */
		}
	
		return rewardofCurrent;
		
	}
        
        //Checks whether the decided job from environment is already in the job queue
        final Job SearchJobinQ(Job the_job)
        {
                
           	int i;
		Job job_in_queue;
        
        	for (i=0;i<m_job_queue.size();i++)
		{
			job_in_queue=(Job)m_job_queue.elementAt(i);
                        if (the_job.target_id == job_in_queue.target_id)
                        {   
                                //if (job_in_queue.job_type==ACTION_SOURCE_CENTER_ENV)
                                //        System.out.println("Combined environment job found.");           
                                if ((job_in_queue.job_type & ACTION_SOURCE_CENTER)>0)
                                {
                                        job_in_queue.job_type=ACTION_SOURCE_CENTER_ENV;
                                        //System.out.println("Environment job same as center job***");
                                }
                                removeFromJobQueue(job_in_queue);
                                return job_in_queue;
                        }
                }
                
                return the_job;

        }

        //Checks whether the dispatcher job is already in the job queue	
        final boolean SearchTargetinQ(int target_id)
        {
                
           	int i;
		Job job_in_queue;
        
        	for (i=0;i<m_job_queue.size();i++)
		{
			job_in_queue=(Job)m_job_queue.elementAt(i);
                        if (target_id == job_in_queue.target_id)  
                        {
                              if ((job_in_queue.job_type & ACTION_SOURCE_ENVIRONMENT)>0)  
                                job_in_queue.job_type = ACTION_SOURCE_CENTER_ENV;         
                              else if ((job_in_queue.job_type & ACTION_SOURCE_CENTER)>0)
                              {
                                m_msgParser.writeMyFile(m_filename,"Center job sent again??");
                              }  
                              return true;
                        }
                }
                return false;
        }
        
	void handleJobPreemption()
	{
	    if (m_current_job==null || ((m_target_index!=WORLD_MODEL_CIVILIAN_INDEX || m_target_status==AMBULANCE_TARGET_STATUS_NOT_FOUND) && (m_target_index!=WORLD_MODEL_ROAD_INDEX || !PoliceFoundTarget())))
            {
                center_job = null;    
                other_alternatives=false;
                
		current_priority_order=m_learner.getPreemptionOrder(m_current_state,m_id);
			
		//SCAN THE JOB QUEUE TO FIND HIGHEST PRIORITY JOB
		for (i=0;i<m_job_queue.size();i++)
		{
			job_in_queue=(Job)m_job_queue.elementAt(i);
			if (m_current_job!=null)
			{
				if (m_current_job.target_id == job_in_queue.target_id)
				{
                                        m_msgParser.writeMyFile(m_filename,"same_job!!! removing!!!");
					removeFromJobQueue(job_in_queue);
					i--;
					continue;
				}
			}
			
			//Take note if there is a center job in queue. This is used to see how many alternatives we have to choose from
			if ((job_in_queue.job_type & ACTION_SOURCE_CENTER)>0)
			{
                                center_job=job_in_queue;
                                if (PRE_LOGGING) m_msgParser.writeMyFile(m_filename,"Job in queue: " + center_job.target_id);
			}
		}
		      
		//WE DO NOT HAVE A JOB, DECIDING ON A TARGET	
		if (m_current_job==null)
		{
                        if (numOfSaveableTargets()>0)
                        {
	              	        other_alternatives=true;
                        }
                        
			if (PRE_LOGGING) m_msgParser.writeMyFile(m_filename,"We don't have a job");
                        
		        priority_of_environment = getPriorityofJob(ACTION_SOURCE_ENVIRONMENT,current_priority_order);
                        if (getPriorityofJob(ACTION_SOURCE_OWN,current_priority_order) > priority_of_environment)
                                priority_of_environment = getPriorityofJob(ACTION_SOURCE_OWN,current_priority_order);
                                
                        priority_of_dispatcher = getPriorityofJob(ACTION_SOURCE_CENTER,current_priority_order);         
		
			if ( center_job!=null && (!other_alternatives || (priority_of_dispatcher>priority_of_environment)))
			{
				
				m_current_job=center_job;
				removeFromJobQueue(center_job);
				/*
                                m_current_job.preempt_hist_index++;
				if (!other_alternatives)
				{
					m_current_job.preempt_hist[m_current_job.preempt_hist_index].recordHistory(TAKE_OUT, m_current_state, ALT_NO_ALT, 0, current_priority_order,m_current_job.job_type);
				}
				//If there is more than one alternative
				else
				{
					m_current_job.preempt_hist[m_current_job.preempt_hist_index].recordHistory(TAKE_OUT, m_current_state, ALT_CENTER_ENV, 0, current_priority_order,m_current_job.job_type);
				}
                                */
				if (PRE_LOGGING) m_msgParser.writeMyFile(m_filename,"Dispatcher job chosen"  + m_current_job.target_id);			
				m_current_job.target_object=(SimObject)m_world_model.getObjectbyId(m_current_job.target_id,m_target_index);	
				
				m_current_job.accumulated_reward=0;
				random_object=null;
                                return;
			}
			else if (other_alternatives && (center_job==null || (priority_of_environment>priority_of_dispatcher)))
			{
				
				m_current_job=decideYourOwnJob(null);
				if (m_current_job!=null)
				{
                                        m_current_job = SearchJobinQ(m_current_job);
					random_object=null;
                                        /*
					m_current_job.preempt_hist_index++;
                                        if (PRE_LOGGING) m_msgParser.writeMyFile(m_filename,"Environment job chosen " + m_current_job.target_id);
					if (center_job==null)
					{
						m_current_job.preempt_hist[m_current_job.preempt_hist_index].recordHistory(TAKE_OUT, m_current_state, ALT_NO_ALT, 0, current_priority_order,m_current_job.job_type);
					}
					//If there are two alternatives
					else
					{
						m_current_job.preempt_hist[m_current_job.preempt_hist_index].recordHistory(TAKE_OUT, m_current_state, ALT_CENTER_ENV, 0, current_priority_order,m_current_job.job_type);  	
					}	
                                        */
				}
				else 
				{
					if (PRE_LOGGING) m_msgParser.writeMyFile(m_filename,"We will ask center for job 1");
                                        m_msgParser.writeMyFile(m_filename,"Env job not decided????????????");
					getaRandomJob();
				}
                                return;
			}
			else
			{
				if (PRE_LOGGING) m_msgParser.writeMyFile(m_filename,"We will ask center for job 2");
				getaRandomJob();
                                return;
			}
		
		}//end of we don't have a job
		
		//WE HAVE A JOB, DECIDING WHETHER TO PREEMPT
                //test case firefighters does not preempt
                //else if ((m_target_index!=WORLD_MODEL_BUILDING_INDEX) && (m_target_index!=WORLD_MODEL_CIVILIAN_INDEX || m_target_status==AMBULANCE_TARGET_STATUS_NOT_FOUND) && (m_target_index!=WORLD_MODEL_ROAD_INDEX || !PoliceFoundTarget()))
		else if ((m_target_index!=WORLD_MODEL_CIVILIAN_INDEX || m_target_status==AMBULANCE_TARGET_STATUS_NOT_FOUND) && (m_target_index!=WORLD_MODEL_ROAD_INDEX || !PoliceFoundTarget()))
		{
			//m_msgParser.writeMyFile(m_filename,"Preemption will start");
			   
                        if (numOfSaveableTargets()>1)     //One saveable target is our own....
                        {
	              	        other_alternatives=true;
                        }   
                           
			//There is no alternative
			if (!other_alternatives && center_job == null)
			{
				//m_msgParser.writeMyFile(m_filename,"There are no other jobs");
				return;
			}
                       
                        priority_of_environment = getPriorityofJob(ACTION_SOURCE_ENVIRONMENT,current_priority_order);
                        priority_of_dispatcher = getPriorityofJob(ACTION_SOURCE_CENTER,current_priority_order);         
                     
                        if (m_current_job.job_type==ACTION_SOURCE_CENTER)
				priority_of_current_job=priority_of_dispatcher;
			else if (m_current_job.job_type==ACTION_SOURCE_ENVIRONMENT)
				priority_of_current_job=getPriorityofJob(ACTION_SOURCE_OWN,current_priority_order); 
			else if (m_current_job.job_type==ACTION_SOURCE_CENTER_ENV)
                        {
                                if (priority_of_dispatcher > priority_of_environment)
                                        priority_of_current_job=priority_of_dispatcher;
                                else
                                        priority_of_current_job=getPriorityofJob(ACTION_SOURCE_OWN,current_priority_order);
                        }
                                                
			if (priority_of_current_job==MAX_PREEMPT_PRIORITY)
			{
				//m_msgParser.writeMyFile(m_filename,"Our job has highest priority " + priority_of_current_job + " target " + m_current_job.target_id + " job type " +  m_current_job.job_type);
				m_current_job.preempt_hist_index++;
                                
				if (m_current_job.job_type==ACTION_SOURCE_ENVIRONMENT)
					m_current_job.preempt_hist[m_current_job.preempt_hist_index].recordHistory(TAKE_OUT, m_current_state, ALT_CENTER_ENV, 0, current_priority_order,ACTION_SOURCE_OWN);  
                                else if (m_current_job.job_type==ACTION_SOURCE_CENTER_ENV)
					m_current_job.preempt_hist[m_current_job.preempt_hist_index].recordHistory(TAKE_OUT, m_current_state, ALT_CENTER_ENV, 0, current_priority_order,ACTION_SOURCE_CENTER_OWN);  
                                else
					m_current_job.preempt_hist[m_current_job.preempt_hist_index].recordHistory(TAKE_OUT, m_current_state, ALT_CENTER_ENV, 0, current_priority_order,m_current_job.job_type);  
				return;
			}

			if (( !other_alternatives || priority_of_current_job>=priority_of_environment) && ( center_job==null || priority_of_current_job>priority_of_dispatcher))
			{
				//m_msgParser.writeMyFile(m_filename,"Our job has the highest priority among alternatives" + priority_of_current_job + " target " + m_current_job.target_id + " job type " +  m_current_job.job_type);
				m_current_job.preempt_hist_index++;
                                
				if (m_current_job.job_type==ACTION_SOURCE_ENVIRONMENT)
					m_current_job.preempt_hist[m_current_job.preempt_hist_index].recordHistory(TAKE_OUT, m_current_state, ALT_CENTER_ENV, 0, current_priority_order,ACTION_SOURCE_OWN);  
				else if (m_current_job.job_type==ACTION_SOURCE_CENTER_ENV)
					m_current_job.preempt_hist[m_current_job.preempt_hist_index].recordHistory(TAKE_OUT, m_current_state, ALT_CENTER_ENV, 0, current_priority_order,ACTION_SOURCE_CENTER_OWN);  

                                else
					m_current_job.preempt_hist[m_current_job.preempt_hist_index].recordHistory(TAKE_OUT, m_current_state, ALT_CENTER_ENV, 0, current_priority_order,m_current_job.job_type);  
			        return;
			}
			else if ( center_job!=null && (!other_alternatives || (priority_of_dispatcher>priority_of_environment)))
			{ 	
				m_current_job.preempt_hist_index++;
			        
				if (m_current_job.job_type==ACTION_SOURCE_ENVIRONMENT)
				{
				     m_current_job.preempt_hist[m_current_job.preempt_hist_index].recordHistory(PUT_BACK, m_current_state, ALT_CENTER_ENV, getStatusOfJob(m_current_job), current_priority_order,ACTION_SOURCE_OWN);  
				}
                                else if (m_current_job.job_type==ACTION_SOURCE_CENTER_ENV)
				{
				     m_current_job.preempt_hist[m_current_job.preempt_hist_index].recordHistory(PUT_BACK, m_current_state, ALT_CENTER_ENV, getStatusOfJob(m_current_job), current_priority_order, ACTION_SOURCE_CENTER_OWN);  
				}
				else
				{
			             m_current_job.preempt_hist[m_current_job.preempt_hist_index].recordHistory(PUT_BACK, m_current_state, ALT_CENTER_ENV, getStatusOfJob(m_current_job), current_priority_order,m_current_job.job_type);  
				}
				m_current_job = preemptJob(center_job, m_current_job);
                                
			        //m_msgParser.writeMyFile(m_filename,"Reverting to center job " + m_current_job.target_id);
				m_current_job.preempt_hist_index++;	
				m_current_job.preempt_hist[m_current_job.preempt_hist_index].recordHistory(TAKE_OUT, m_current_state, ALT_CENTER_ENV, 0, current_priority_order,m_current_job.job_type);  
				m_current_job.accumulated_reward=0;
				random_object=null;
                                return;
			
			}
			else if (other_alternatives && (center_job==null || (priority_of_environment>priority_of_dispatcher)))
			{
				//FIND A JOB FROM ENVIRONMENT, NOT YOUR CURRENT ONE	
				Job job_from_environment;
                                //29102002 Changing this line. Real environment choice is to be done
				//job_from_environment=decideJobfromAround(m_current_job);

                                job_from_environment=decideYourOwnJob(m_current_job);
                                
				if (job_from_environment!= null)
				{
                                        //Environment job cannot preempt environment job thus the job should not be taken out recently and type environment
                                        if( !( m_current_job.preempt_hist_index>=0 && (m_current_job.preempt_hist[m_current_job.preempt_hist_index].job_type & ACTION_SOURCE_ENVIRONMENT)>0 && m_current_job.preempt_hist[m_current_job.preempt_hist_index].preemption_direction==TAKE_OUT))
                                        {
					   m_current_job.preempt_hist_index++;
                                        
					   if (m_current_job.job_type==ACTION_SOURCE_ENVIRONMENT)
				           {
				                 m_current_job.preempt_hist[m_current_job.preempt_hist_index].recordHistory(PUT_BACK, m_current_state, ALT_CENTER_ENV, getStatusOfJob(m_current_job), current_priority_order,ACTION_SOURCE_OWN);  
				           }
                                           else if (m_current_job.job_type==ACTION_SOURCE_CENTER_ENV)
				           {
				                 m_current_job.preempt_hist[m_current_job.preempt_hist_index].recordHistory(PUT_BACK, m_current_state, ALT_CENTER_ENV, getStatusOfJob(m_current_job), current_priority_order, ACTION_SOURCE_CENTER_OWN);  
			             	   }
			             	   else
			             	   {
			                        m_current_job.preempt_hist[m_current_job.preempt_hist_index].recordHistory(PUT_BACK, m_current_state, ALT_CENTER_ENV, getStatusOfJob(m_current_job), current_priority_order,m_current_job.job_type);  
			             	   }
                                        }   
					putToJobQueue(m_current_job);
                                        m_current_job = SearchJobinQ(job_from_environment);
                                        //m_msgParser.writeMyFile(m_filename,"Reverting to environment job " + m_current_job.target_id);
					
					m_current_job.preempt_hist_index++;
					m_current_job.preempt_hist[m_current_job.preempt_hist_index].recordHistory(TAKE_OUT, m_current_state, ALT_CENTER_ENV, 0, current_priority_order,m_current_job.job_type);  
                                        m_current_job.accumulated_reward=0;
					random_object=null;	
				}
                                return;
			}
		}//we have a job	   
            }//End of initial check
            //This log is tested and found to be useful.
            //else
            //{
            //    System.out.println("Preemption not done " + m_id);
            //}
	}//end of handlepreemption
	
	boolean PoliceFoundTarget()
	{
		if (PRE_LOGGING) m_msgParser.writeMyFile(m_filename,"super's PoliceFoundTarget called" );
		return false;
	}
	
        
	void getaRandomJob()
	{//Random job is a road for a police officer, a building for a ff or ambulance
	
		byte targetType = 0;
                
		//Checking whether previous random target has been arrived to
		if (random_object!=null)
		{
                        //Not at or adjacent to the random target
			if (!(m_world_model.notAdjacent (random_object.m_id, we_are_at))) 
			{		
				//A building may get fire any time, so for firefighters it does not matter if a target has been visited or not
				if (m_target_index!=WORLD_MODEL_BUILDING_INDEX) m_excluded_targets.add(new Integer(random_object.m_id));			
				random_object=null;
			}
			
			if (random_object!=null)
				if (PRE_LOGGING)  m_msgParser.writeMyFile(m_filename,"Random target kept: "+ random_object.m_id);	
			else
				if (PRE_LOGGING)  m_msgParser.writeMyFile(m_filename,"Random target discarded");	
		}
		
		//If there were no random target, or we have arrived to previous random target
		if (random_object==null)
		{
			if (m_target_index==WORLD_MODEL_BUILDING_INDEX)
                                //FF 1: Finished jobs icinde olmayan binalara gidiyor
				random_object = m_world_model.findRealAny(m_world_model.the_world[m_target_index],we_are_at,m_excluded_targets);	
			else if (m_target_index==WORLD_MODEL_CIVILIAN_INDEX)
		       	{
	                        //AMB 1: Injured civilian locationlara gidiyor		
				random_object=m_world_model.findClosest(injured_civilian_locations,we_are_at,false);
				targetType=1;	
				if (random_object==null)
				{
                                        //AMB 2: Visited random harici yollara gidiyor
					random_object = m_world_model.findRealAny(m_world_model.the_world[WORLD_MODEL_BUILDING_INDEX],we_are_at,m_excluded_targets);		
				        targetType=2;
                                }
				else
				{
					injured_civilian_locations.remove(new Integer (random_object.m_id));
					//System.out.println("ambulance is going to a heard civilian location");
                                        targetType=3;
				}
			
			}
			else	
			{
				takeActionRandomArrived();		
			        if (!current_action_done)
                                {
                                        //POL 1: Closest road that is saveable
				    random_object = m_world_model.findClosest(m_world_model.the_world[WORLD_MODEL_ROAD_INDEX],we_are_at,true);
                                        targetType=1;
                                
				    if (random_object==null)
				    {
                                        //POL 2: Visited random harici burning buildinge gidiyor
					random_object = m_world_model.findClosestExcept(m_world_model.the_world[WORLD_MODEL_BUILDING_INDEX],we_are_at,true,m_excluded_targets);
					if (random_object==null)	
                                                //POL 3: Visited random harici refugelere gidiyor
						{random_object = m_world_model.findClosestExcept(m_world_model.the_world[WORLD_MODEL_REFUGE_INDEX],we_are_at,false,m_excluded_targets);	
                                                 targetType=2;}
					if (random_object==null)
					        //POL 4: Visited random harici yollara gidiyor
						{random_object = m_world_model.findRealAny(m_world_model.the_world[WORLD_MODEL_ROAD_INDEX],we_are_at,m_excluded_targets);
                                                 targetType=3;}       		
				    }
                                }
			}
			
                        if (random_object!=null)
			     if (PRE_LOGGING)  m_msgParser.writeMyFile(m_filename,"Random target decided: "+ random_object.m_id + " type: " + targetType);	
			else
                             if (PRE_LOGGING)  m_msgParser.writeMyFile(m_filename,"Random target could not be decided.");	   
			//When we decideon a new random target, we ask the kernel for job
			if (withinSayLimits())
			{
				if (COMM_LOGGING) m_msgParser.writeMyFile(m_filename,"SENT: job request");
				m_io.sendAK_TELL(m_id, MSG_REQUEST, m_myself.get_position());   
				increaseSaidNum();
			}	
		}
											
		if (random_object!=null  && m_myself.get_buriedness()==0) 
		{
			
			//Road targets are indicated directly including blocked endpoint
			//Building targets are sent with removelast since an agent should not get blocked in a building

		        if (m_agent_index== WORLD_MODEL_POLICE_INDEX )
                        {
                                takeActionRandom();
                        }
                        else
                        {
	                        if (random_object instanceof Building)
		              		moveTo(random_object.m_id, true);
	              		else if (!(random_object instanceof MovingObject))  //Target road or node
        				moveTo(random_object.m_id, false);
		              	else
	               	      	{
	               			random_object=null;
	               			if (PRE_LOGGING)  m_msgParser.writeMyFile(m_filename,"Random could not send move to");
	               		}
                        }
		}
						

	}//end of random job 
	
	Job decideYourOwnJob(Job current_job)
	{
		Job decidedJob;
		byte decided_action;
		SimObject decided_target=null;
	
		if (numOfSaveableTargets()==1)
		{
                        if (current_job != null) return null; //There is only one job, we cannot change it                        
                        
			
			decided_target=(SimObject)m_world_model.findOnly();
			if (decided_target!=null)
			{
				decidedJob=new Job(decided_target.m_id,ACTION_SOURCE_ENVIRONMENT);
				decidedJob.q_start_state=m_current_state;
                                decided_target.m_assignedTo=1; //In order not to report own job as an injured civilian
				decidedJob.target_object=decided_target;
                                //m_msgParser.writeMyFile(m_filename,"Only job returned as environment job " +  decided_target.m_id);
				return decidedJob;
			}
                        else
                        { 
                                m_msgParser.writeMyFile(m_filename,"Findonly returns null??");
                                return null;
                        }
		}
		else{
		decided_action = m_learner.decideOnAction(m_current_state);		
		
		switch (decided_action)
		{		
			case WORKER_URGENT_FIRST:
                                //CLOSEST OF THESE SHOULD BE RETURNED
				decided_target = findMaxSeverity( m_x, m_y, m_current_time,current_job);
                                if (decided_target != null)
                                {
                                        if (PRE_LOGGING) m_msgParser.writeMyFile(m_filename,"Urgent job is: " + decided_target.m_id);
                                }
                                else
                                {
                                        m_msgParser.writeMyFile(m_filename,"Urgent job returned null");
                                }
				break;
			
			case WORKER_CLOSEST_FIRST:
				if (m_agent_index==WORLD_MODEL_FIREFIGHTER_INDEX) 
				{       //Lightest of the closest is returned    
					decided_target = m_world_model.findMinSevere_buildingAround((current_job == null ? 0:current_job.target_id), m_x,m_y);

                                        if (decided_target == null)
                                                decided_target = m_world_model.FindClosestBurningBuilding(m_myself,MY_MAXSHORT,current_job);   
                                }
			        else
				{
					decided_target = m_world_model.findClosestTarget(m_myself,current_job);   
                                }
                                if (decided_target != null)
                                {
                                        if (PRE_LOGGING) m_msgParser.writeMyFile(m_filename,"Closest job is: " + decided_target.m_id);
                                       
                                }
                                else
                                {
                                        m_msgParser.writeMyFile(m_filename,"Closest job returned null");
                                }
                                
				break;
			
			case WORKER_LIGHTEST_FIRST:
                                //CLOSEST OF THESE SHOULD BE RETURNED
				decided_target = m_world_model.findMinSeverity(m_target_index, m_x, m_y, m_current_time,current_job);
                                if (decided_target != null)
                                {       
                                        if (PRE_LOGGING) m_msgParser.writeMyFile(m_filename,"Lightest job is: " + decided_target.m_id);
                                }
                                else
                                {
                                        m_msgParser.writeMyFile(m_filename,"Lightest job returned null");
                                }
				break;
							
			default:
				System.out.println("Error in AgentThd.java. No job could be found.");
				decided_target = null;
				break;
		}//end of switch
		
		
		if (decided_target!=null)
		{
			decidedJob=new Job (decided_target.get_id(),ACTION_SOURCE_ENVIRONMENT,m_current_state,decided_action);
                        decided_target.m_assignedTo=1; //In order not to report own job as an injured civilian
			decidedJob.target_object=decided_target;
                        //m_msgParser.writeMyFile(m_filename,"Target " +  decided_target.m_id + " decision by " + decided_action);
			return decidedJob;
		}
                else
                {
                        return null;
                }
		
		}//end of one job only control

		//return null;
	
	}//end of find your own job
	
	/*void discoverExecuteMessage()
	{
		Integer temp_index;
		int execute_message=m_execute_base;
		int minimum_agent_id=MY_MAXINT;
		int i;

		Iterator it=m_world_model.the_world[m_agent_index].keySet().iterator();

		while(it.hasNext())
		{
			temp_index=(Integer)it.next();

			if (temp_index.intValue()<minimum_agent_id)
				minimum_agent_id=temp_index.intValue();
		}


		for (i=minimum_agent_id+1; i<=m_id;i++)
		{
			if (m_world_model.the_world[m_agent_index].containsKey(new Integer(i)))
				execute_message++;
		}

		m_myself.m_execute_message=execute_message;

	}// end of discoverExecuteMessage*/

}//end of AgentThd

class PreemptionHistory implements Definitions
{
	byte preemption_direction;     //PUT_BACK or TAKE_OUT
	byte [] preemption_order;
	int preemption_state;	
	byte preemption_alternative;   //Is it the only one
	double status_when_put_in_Q;
	int job_type;                  //This is used to handle     ENV job to OWN job transitions
	
	PreemptionHistory()
	{
		preemption_direction=-1;
		preemption_state=-1;
		preemption_alternative=-1;
		status_when_put_in_Q=0;
		preemption_order=new byte[def_action_source_num];
		preemption_order[0]=-1;
		preemption_order[1]=-1;
		preemption_order[2]=-1;
		job_type=-1;
	}
	
	void recordHistory(byte x_direction, int x_state, byte x_alternative, double x_status, byte [] x_order, int x_job_type)
	{
		preemption_direction=x_direction;
		preemption_state=x_state;
		preemption_alternative=x_alternative;
		status_when_put_in_Q=x_status;
		preemption_order=(byte [])x_order.clone();
		job_type = x_job_type;
	}
}

class Job implements Definitions
{
	int target_id;		           //The id of the target we are going for
	byte job_type;				   //Own, center, environment
	int q_start_state;			   //The state we decided on the target
	byte q_start_action;			   //The action that yielded the target(own decision)
	boolean took_action;   
	SimObject target_object;
	int cycles_spent_on_job;                 //Used in police time discount
	double accumulated_reward; 	           //reward occuring because of jobs in queue being saved or ruined
	PreemptionHistory [] preempt_hist;
	int preempt_hist_index;		   //Points to the last preemption history record
	double initial_status;

	Job(int x_target_id, byte x_job_type)
	{
		target_id=x_target_id;
		job_type=x_job_type;
		cycles_spent_on_job=0;
		accumulated_reward=0;	
		q_start_state=-1;		
		q_start_action=-1;		
		took_action=false;
		preempt_hist=new PreemptionHistory[DEF_NUM_TICKS];
		preempt_hist_index=-1;
		initial_status=0;
                
		int i;
		for (i=0;i<DEF_NUM_TICKS;i++)
		{
			preempt_hist[i]=new PreemptionHistory();
		}
	}
	
	Job(int x_target_id, byte x_job_type, int x_q_start_state, byte x_q_start_action)
	{
		target_id=x_target_id;
		job_type=x_job_type;
		took_action=false;
		cycles_spent_on_job=0;
		accumulated_reward=0;
		preempt_hist=new PreemptionHistory[DEF_NUM_TICKS];
		preempt_hist_index=-1;
                q_start_state=x_q_start_state;		
		q_start_action=x_q_start_action;
		initial_status=0;
                took_action=false;
		
		int i;
		for (i=0;i<DEF_NUM_TICKS;i++)
		{
			preempt_hist[i]=new PreemptionHistory();
		}	
	}
 
}
