class AmbulanceTeam extends AgentThd
{

	int m_refuge_id;
	int targetPosition;

	AmbulanceTeam(int agent_type, String prefix, String inetaddr, int training_run_num)
	{
		super(agent_type,prefix,inetaddr, training_run_num);
		m_target_status=AMBULANCE_TARGET_STATUS_NOT_FOUND;
		m_refuge_id=-1;
	}


	final void SetAgentSpecificConstants()
	{	//These indexes are for indexing the world model array, not related to kernel
		m_agent_index=WORLD_MODEL_AMBULANCE_INDEX;
		m_target_index=WORLD_MODEL_CIVILIAN_INDEX;
		m_execute_base=MSG_AMBULANCE_EXECUTE;
		//m_timeout=DEF_AMBULANCE_TIME_OUT;
		m_positive_reward=AMBULANCE_REWARD;
	}

	
	final boolean ArrivedToRefuge()
	{

		if ( m_myself.get_position()==m_refuge_id )
			return true;
		else
			return false;
	}
	
	final boolean CivilianArrivedToRefuge()
	{

		if (m_current_job.target_object == null)
		{
			if (AMB_LOGGING) m_msgParser.writeMyFile(m_filename,"CivilianArrivedToRefuge: Target object is null.");
			return false;
		}

		//Target position is re-set here since this function is called independent of takeAction
		targetPosition=((MovingObject)m_current_job.target_object).get_position();

		if (AMB_LOGGING) m_msgParser.writeMyFile(m_filename,"CivilianArrivedToRefuge:Civilian is at. " + targetPosition + " target id " + ((MovingObject)m_current_job.target_object).get_id());
		if (m_world_model.the_world[WORLD_MODEL_REFUGE_INDEX].containsKey(new Integer(targetPosition)))
			return true;
		else
			return false;
	}

	final boolean BuriedTarget()
	{
		if (m_world_model.targetIsHere(m_current_job.target_id) && ((MovingObject)m_current_job.target_object).get_buriedness()>0)
		{
			if (AMB_LOGGING){ 
				m_msgParser.writeMyFile(m_filename,"Civilian buriedness: " + ((MovingObject)m_current_job.target_object).get_buriedness()); 
				m_msgParser.writeMyFile(m_filename,"Civilian healthiness: " + ((MovingObject)m_current_job.target_object).get_hp()); 
				m_msgParser.writeMyFile(m_filename,"Civilian damage: " + ((MovingObject)m_current_job.target_object).get_damage()); 
			}
			return true;
		}
		else
			return false;
	}

	final byte targetRuined(Job this_job)
	{
		if (this_job.target_object==null)
		{
		
			return JOB_STATUS_UNKNOWN;
		}
		
		if (m_world_model.dead(this_job.target_object))
		{
			return RUINED;
		}

		return NOT_RUINED;
	
	}

	final byte targetSaved()
	{
		
		if (m_current_job.target_object==null)
		{
			return JOB_STATUS_UNKNOWN;
		}
		if (CivilianArrivedToRefuge()  && !(m_world_model.dead(m_current_job .target_object)))
		{
			if ( m_current_job.took_action  )
			{
				return SAVED_BY_ME;
			}
			else
			{
				return SAVED_BY_OTHER;
			}
		}
		return NOT_SAVED_YET;
	}
	
	
	final boolean targetInSaveableRange()
	{
		int i;
		int [] edges;
		
		
		if (m_world_model.targetIsHere(m_current_job.target_id))
		{
			 if (m_myself.get_position()==targetPosition)
				return true;
				
			 if (NEAR_AGENT_RESCUABLE)
			 {
				//If we are on a node and target is on a road adjacent to the node return true
				if (we_are_at instanceof Node)
				{
					edges=((Node)we_are_at).get_edges();
					for(i=0;i<edges.length;i++)
					{
						if (edges[i]==targetPosition && m_world_model.isRoad(targetPosition)) 
						{
							if (AMB_LOGGING) m_msgParser.writeMyFile(m_filename,"We are at node, target in road");
							return true;
						}
					}
				}
				//If we are on a road and target is on a neighbor node we can save it
				else if (we_are_at instanceof Road)
				{
					if(targetPosition == ((Road)we_are_at).get_head())
					{
						if (AMB_LOGGING) m_msgParser.writeMyFile(m_filename,"We are at road, target in node");
						return true;
					}
					if(targetPosition == ((Road)we_are_at).get_tail())
					{
						if (AMB_LOGGING) m_msgParser.writeMyFile(m_filename,"We are at road, target in node");
						return true;
					}
				}
			 }
		}
		return false;	
	}
	
	final boolean targetInRangeSpecial()
	{

		int i;
		int [] edges;
		
			 if (m_myself.get_position()==targetPosition)
				return true;
				
			 if (NEAR_AGENT_RESCUABLE)
			 {
				//If we are on a node and target is on a road adjacent to the node return true
				if (we_are_at instanceof Node)
				{
					edges=((Node)we_are_at).get_edges();
					for(i=0;i<edges.length;i++)
					{
						if (edges[i]==targetPosition && m_world_model.isRoad(targetPosition)) 
						{
							if (AMB_LOGGING) m_msgParser.writeMyFile(m_filename,"We are at node, target in road");
							return true;
						}
					}
				}
				//If we are on a road and target is on a neighbor node we can save it
				else if (we_are_at instanceof Road)
				{
					if(targetPosition == ((Road)we_are_at).get_head())
					{
						if (AMB_LOGGING) m_msgParser.writeMyFile(m_filename,"We are at road, target in node");
						return true;
					}
					if(targetPosition == ((Road)we_are_at).get_tail())
					{
						if (AMB_LOGGING) m_msgParser.writeMyFile(m_filename,"We are at road, target in node");
						return true;
					}
				}
			 }
		return false;	
	}
	
	
	final boolean targetInAmbulance()
	{
		
		return (m_world_model.targetIsHere(m_current_job.target_id) && (m_id==targetPosition));
	
	}

	final void takeAction()
	//AMBULANCE TARGET IS CIVILIAN (CURRENTLY)
	{//If the target is OK or lost this is determined beforehand, thus this function only deals with the 'not arrived to target' and 'arrived to target but not finished the job' cases
	
	
	if (!current_action_done && m_myself.get_buriedness()==0 && m_target_status!=AMBULANCE_TARGET_STATUS_UNLOADED )
	  {
		int direction;
		int job_id=m_current_job.target_id;
		targetPosition=	((MovingObject)(m_current_job.target_object)).get_position();
		//m_current_job.cycles_spent_on_job++;
		
		//System.out.println("Ambulance has a target...");

		//We think we arrived to target, but the target is not at the position we think it is
		if (!m_world_model.targetIsHere(m_current_job.target_id) && targetInRangeSpecial())
		{
			if (AMB_LOGGING) m_msgParser.writeMyFile(m_filename,"TARGET POSITION WRONG");
			((MovingObject)m_current_job.target_object).m_position=m_world_model.findDummyRefuge();
			if (AMB_LOGGING) m_msgParser.writeMyFile(m_filename,"Civilian yok oldu.-2");										  	EndOfJob();
			resetJobIndicators(job_id);
			return;
		}

		if (targetPosition != m_id)
		{
			Integer targetPosInt=new Integer(targetPosition);
			if (m_world_model.the_world[WORLD_MODEL_AMBULANCE_INDEX].containsKey(targetPosInt) || m_world_model.the_world[WORLD_MODEL_REFUGE_INDEX].containsKey(targetPosInt))
			{
				if (AMB_LOGGING) m_msgParser.writeMyFile(m_filename,"Our target is handled by someone else/or moved by itself when rescued...");
				((MovingObject)m_current_job.target_object).m_position=m_world_model.findDummyRefuge();
					if (AMB_LOGGING) m_msgParser.writeMyFile(m_filename,"Civilian yok oldu.-1");										  EndOfJob();
					resetJobIndicators(job_id);
				return;
			} 
		}
		else if (m_target_status!=AMBULANCE_TARGET_STATUS_RESCUING && m_target_status!=AMBULANCE_TARGET_STATUS_LOADED)
		{
			System.out.println("Something wrong with the ambulance. We haven't loaded a civilian but it is there!!!");
			if (AMB_LOGGING) m_msgParser.writeMyFile(m_filename,"Something wrong with the ambulance. We haven't loaded a civilian but it is there!!!");
		}
		if (AMB_LOGGING) m_msgParser.writeMyFile(m_filename,"Ambulance takeaction target:" + m_current_job.target_id + " target pos " + targetPosition + " time  " + m_current_time);

		if (m_target_status==AMBULANCE_TARGET_STATUS_NOT_FOUND)
		{
			if (!targetInSaveableRange())
			{
				//System.out.println("1.Far from target.");
				if (AMB_LOGGING) m_msgParser.writeMyFile(m_filename,"1.Far from target.");
				m_target_status=AMBULANCE_TARGET_STATUS_NOT_FOUND;
				moveTo(targetPosition,false);		//We shall not move to the Civilian but where the civilian stands at
			}
			else if (targetInSaveableRange() && m_world_model.saveable(m_current_job.target_object))
			{
				//System.out.println("2.Targeti bulduk rescue edicez");
				if (AMB_LOGGING) m_msgParser.writeMyFile(m_filename,"2.Targeti bulduk rescue edicez");
				m_current_job.took_action=true;
				m_target_status=AMBULANCE_TARGET_STATUS_RESCUING;
			}
			else	//The object has moved somewhere or needs not be saved
			{
				if (AMB_LOGGING) m_msgParser.writeMyFile(m_filename,"Target does not need to be saved.");
				m_target_status=AMBULANCE_TARGET_STATUS_NOT_FOUND;
				//WE FAIL!!!
				handleCenterReply(m_current_job, MSG_NEUTRAL_POS);
				resetJobIndicators(m_current_job.target_id);
				return;
			}
		}

		if(m_target_status==AMBULANCE_TARGET_STATUS_RESCUING)
		{
			if (BuriedTarget())
			{
				if ( !targetInSaveableRange())
				{
					((MovingObject)m_current_job.target_object).m_position=m_world_model.findDummyRefuge();
					if (AMB_LOGGING) m_msgParser.writeMyFile(m_filename,"Civilian yok oldu.0");
					EndOfJob();
					return;
				}
				//System.out.println("3.Targetimiz buriedmis.");
				//if (AMB_LOGGING) m_msgParser.writeMyFile(m_filename,"3.Targetimiz buriedmis.");
				//System.out.println("Ambulance rescue yolladi");
				m_io.send_two_int_msg(AK_RESCUE,m_id, m_current_job.target_id);
				current_action_done=true;
				m_current_job.took_action=true;
			}
			else
			{
				if (AMB_LOGGING) m_msgParser.writeMyFile(m_filename,"4.Buried degilmis.");
				//System.out.println("4.Buried degilmis.");
				//If target has moved outside our visible range or target is not in our saveable range
				//We should check first condition first, since if a target moves out of range, its location
				//property is the old one

				if (!targetInSaveableRange() && !targetInAmbulance() )
				{
					((MovingObject)m_current_job.target_object).m_position=m_world_model.findDummyRefuge();
					if (AMB_LOGGING) m_msgParser.writeMyFile(m_filename,"Civilian yok oldu.1");
					EndOfJob();
					return;
				}

				//If we have not loaded the target
				if (AMB_LOGGING) m_msgParser.writeMyFile(m_filename,"Target obj position: " + targetPosition);

				if (!targetInAmbulance())
				{
					//System.out.println("Load yolladik.");
					m_io.send_two_int_msg(AK_LOAD,m_id, m_current_job.target_id);
					if (AMB_LOGGING) m_msgParser.writeMyFile(m_filename,"Load yolladik.");
					current_action_done=true;
					m_current_job.took_action=true;
				}
				else
				{
					m_target_status=AMBULANCE_TARGET_STATUS_LOADED;
					if (AMB_LOGGING) m_msgParser.writeMyFile(m_filename,"Load etmisiz.");
				}
			}

		}

		if (m_target_status==AMBULANCE_TARGET_STATUS_LOADED)
		{
			if (AMB_LOGGING) m_msgParser.writeMyFile(m_filename,"5.Civiliani load ettik refuge ariyoz.");
			
			//If target has moved outside our visible range or target is not in ambulance
			//We should check first condition first, since if a target moves out of range, its location 
			//property is the old one
			if (!targetInAmbulance() )
				{
					((MovingObject)m_current_job.target_object).m_position=m_world_model.findDummyRefuge();
					if (AMB_LOGGING) m_msgParser.writeMyFile(m_filename,"Civilian yok oldu.2");
					EndOfJob();
					return;
				}
				
			if (m_refuge_id==-1)
			{
				SimObject m_refuge;
				m_refuge=m_world_model.findClosest(m_world_model.the_world[WORLD_MODEL_REFUGE_INDEX],we_are_at,false);
				m_refuge_id=m_refuge.get_id();
				if (AMB_LOGGING) m_msgParser.writeMyFile(m_filename,"Refuge id: " + m_refuge_id);
				moveTo(m_refuge_id,false);
			}
			else
			{
				if (ArrivedToRefuge())
				{
					m_io.send_one_int_msg(AK_UNLOAD,m_id);
					if (AMB_LOGGING) m_msgParser.writeMyFile(m_filename,"6.Unload yolladik.");
					current_action_done=true;
                                        m_refuge_id=-1;
				}
				else
				{
					moveTo(m_refuge_id,false);
				}
			}
		}
	  }//Current action done test.	
	}
        
        final void takeActionRandom()
	{
        }
        
        final void takeActionRandomArrived()
        {}
        
        final SimObject findMaxSeverity(int m_x, int m_y, int current_time, Job current_target)
        {
                return m_world_model.findMaxSeverity_civilian(current_target);
        }
        
        
        final double getStatusOfJob(Job theJob)
        {
                return ((Civilian)theJob.target_object).get_hp();
        }
        
        final void setStatusOfJob(Job theJob)
        {
                theJob.initial_status=((Civilian)theJob.target_object).get_hp();
        }
        
        final double compareStatuswithOld(Job theJob)
        {
                        if (m_world_model.dead(theJob.target_object))
			{	
				handleCenterReply(theJob, MSG_NEUTRAL_NEG);
				removeFromJobQueue(theJob);
				return -QUEUE_REWARD;
			}	
			//If the civilian is not dead but not saveable, either another ambulance has got it or it is in a refuge...	
			else if (!m_world_model.saveable(theJob.target_object))
			{
				handleCenterReply(theJob, MSG_NEUTRAL_POS);
				removeFromJobQueue(theJob);
				return QUEUE_REWARD;
			}
                        
                        return 0.0;
        }
        
        //When this function is called theJob is already known to be unsaveable
        final double compareStatuswithOldUnsaveable(Job theJob)
        {
                        if (m_world_model.dead(theJob.target_object))
			{	
				handleCenterReply(theJob, MSG_NEUTRAL_NEG);
				removeFromJobQueue(theJob);
				return -ACCUMULATED_REWARD;
			}	
			//If the civilian is not dead but not saveable, either another ambulance has got it or it is in a refuge
			else  //if ! saveable condition already satisfied
			{
				handleCenterReply(theJob, MSG_NEUTRAL_POS);
				removeFromJobQueue(theJob);
				return ACCUMULATED_REWARD;
			}
                      
        }
        
}
