import java.io.*;
import java.net.*;
import java.util.*;

class IO implements Definitions{

	InetAddress sendAddress;
	int sendPort;
	int receivePort;	
	DatagramSocket ioSocket;
	byte[] connectionBuffer=new byte [DEF_IO_BUFFER_SIZE];
	static short LongUDPID=0;


	//Creates a datagram socket, sets the initial address and port for sending messages to kernel
	IO(InetAddress address, int port)
	{
		try{
			sendAddress=address;
			sendPort=port;
			ioSocket=new DatagramSocket();
		}catch (Exception e){
			System.err.println("IO constructor: Cannot create datagram socket.");
			System.exit(1);
		}
	}



	//Connects to kernel, gets ID and initial world information,returns -1 if any of these steps fail
	final int connectToKernel(int agent_type,WorldModel agents_world)
	{
		int agent_id=-1;	
		
		if (!sendAK_CONNECT(agent_type)) 
			{
				System.out.println("IO connectToKernel : Cannot send AK_CONNECT");
				return -1;
			}
		agent_id=receiveKA_CONNECT_OK(agents_world);
		if (agent_id==-1)
			{
				System.out.println("IO connectToKernel : Cannot get KA_CONNECT_OK");
				return -1;
			}
		if (!sendAK_ACKNOWLEDGE(agent_id))
			{
				System.out.println("IO connectToKernel : Cannot send AK_ACKNOWLEDGE");
				return -1;
			}

		System.out.println("Connected to kernel, ID :" + agent_id);

		return agent_id;
	}

	//This function is redundant now
	//Returns a raw byte [] containing the 8 byte header, returns null if packet could not be received
	final byte []	receivePacket()
	{
		byte [] returnArr;

		DatagramPacket rcvPacket=new DatagramPacket(connectionBuffer, connectionBuffer.length);
		try{
			ioSocket.receive(rcvPacket);
			returnArr =rcvPacket.getData();
		}catch (Exception e){
			System.out.println("IO receivePacket: Could not receive packet from kernel.");
			returnArr=null;
		}
		
		return returnArr;
	}


	final int [] receiveLongPacket()
	//This returns an int array without the 8 byte header of the packet, goes like type,size,body,returns null if a message could not be received
	{
	try {
	
		DatagramPacket rcvPacket;
		Hashtable messagePool;			//Hashtable for all messages with different ids
		MessageSorter messageSorter;	//Data structure for sorting all messages with same id
		MessageParser messageParser;	//Message parser class
		byte [] singlePacket;
		int []  singlePacketBody;
		int [] combinedPacket;
		int packetID=-1;
		int packetCount=0;
		int packetTotal=0;
		int i;
		
		messagePool=new Hashtable();
		messageParser=new MessageParser();

		while (true){

		try{
			//RECEIVE PACKET
			rcvPacket=new DatagramPacket(connectionBuffer, connectionBuffer.length);
			ioSocket.receive(rcvPacket);
			singlePacket =rcvPacket.getData();
		}
		catch(Exception e)
		{
			System.out.println("IO receiveLongPakcet: Could not receive long UDP packet.");
			return null;
		}
		/*
		if (rcvPacket.getLength()<40)
		{
			System.out.println("The message just received is");
			int k;
			for (k=0;k<rcvPacket.getLength();k++)
				System.out.println( k + ": " + singlePacket[k]);
		}
		*/
		packetID= messageParser.byteArrtoInt (singlePacket, MSG_HEADER_ID_INDEX, 2);

		//We are receiving a completely new message, we will open a hashtable entry
		if (!messagePool.containsKey(new Integer(packetID)))
		{
			packetTotal= messageParser.byteArrtoInt (singlePacket, MSG_HEADER_TOTAL_INDEX, 2);
			
			//If KA_SENSE or KA_CONNECT_OK were not so long after all, we will send the only packet
			if (packetTotal==1)
			{	
					
				singlePacketBody=messageParser.byteArrtoIntArr (singlePacket, MSG_FORMAT_TYPE_INDEX, rcvPacket.getLength()-MSG_FORMAT_TYPE_INDEX);
				return singlePacketBody;
			}
			packetCount= messageParser.byteArrtoInt (singlePacket, MSG_HEADER_COUNT_INDEX, 2);
			
			if (packetCount<0 || packetCount>=packetTotal)
			{
				System.out.println("IO receiveLongPakcet: problem with packetcount: " + packetCount + "packetTotal: "+ packetTotal);
				return null;
			}
			
			messageSorter=new MessageSorter();
			messageSorter.id=packetID;
			messageSorter.totalCount=packetTotal;
			messageSorter.rcvCount=1;
			messageSorter.messages= new int [packetTotal][];
			for(i=0;i<packetTotal;i++)
				messageSorter.messages[i]=null;

			singlePacketBody=messageParser.byteArrtoIntArr (singlePacket, MSG_FORMAT_TYPE_INDEX, rcvPacket.getLength()-MSG_FORMAT_TYPE_INDEX);

			messageSorter.messages[packetCount]=singlePacketBody;
			messagePool.put(new Integer(packetID),messageSorter);

		}

		//If we have received another packet of this message before
		else
		{
			packetCount= messageParser.byteArrtoInt (singlePacket, MSG_HEADER_COUNT_INDEX, 2);
			messageSorter=(MessageSorter)messagePool.get(new Integer(packetID));
			
			//If we are receiving this packet for the first time
			if (messageSorter.messages[packetCount]==null)
			{

				singlePacketBody=messageParser.byteArrtoIntArr (singlePacket, MSG_FORMAT_TYPE_INDEX, rcvPacket.getLength()-MSG_FORMAT_TYPE_INDEX);

				messageSorter.messages[packetCount]=singlePacketBody;
				messageSorter.rcvCount++;
				
				//If all packets of this message are complete,combine the packets
				if (messageSorter.rcvCount>=messageSorter.totalCount)
				{
					combinedPacket=messageParser.combinePackets(messageSorter.messages);
					messagePool.remove(new Integer(packetID));

					//If we have received a connect ok message, we update the address and port to the origin of this message
					if (messageParser.get_msg_type_long_packet(combinedPacket)==KA_CONNECT_OK)
					{
						sendAddress=rcvPacket.getAddress();
						sendPort=rcvPacket.getPort();
					}
					return combinedPacket;
				}//end of packets complete

			}//end of add new packet			
			
		}//end of add packet of existing message

		}//end of while loop
		
	}
	catch (Exception e)
	{
		return null;
	
	}
	}

	final boolean sendPacket(byte [] sendData)
	//Sends a datagram packet to kernel,returns false if unsuccessful
	{
	/*
		if (sendData.length<40)
		{
			System.out.println("SENT:");
			int k;
			for (k=0;k<sendData.length;k++)
				System.out.println( k + ": " + sendData[k]);
		}
	*/
		try{
			DatagramPacket sendPacket=new DatagramPacket(sendData, sendData.length,sendAddress,sendPort);
			ioSocket.send(sendPacket);
			return true;
		}catch (Exception e){

			System.out.println("IO sendPacket: Cannot send packet.");
			return false;
		}
	}


	//Combines the header and body of message, returns null if unsuccessful
	final byte [] preparePacket(int msg_type, byte[] msg_body )
	{	
		ByteArrayOutputStream baos= new ByteArrayOutputStream();
		DataOutputStream dos= new DataOutputStream(baos);

		try{
			dos.writeShort(0x0008);  		//magic number
			dos.writeShort(LongUDPID++);	//id of this udp packet
			dos.writeShort(0x0000);  		//this is the first pakcet
			dos.writeShort(0x0001);			//there is only one packet
		
			dos.writeInt(msg_type);         //AK_SAY, AK_MOVE etc.
			dos.writeInt(msg_body.length);
			//System.out.println("sending msg of size" + msg_body.length);
			dos.write(msg_body,0,msg_body.length);

			dos.writeInt(HEADER_NULL);			//pointer to next packet:NULL  	
			return baos.toByteArray();
		
		}catch(Exception e){
			System.out.println("IO preparePacket: Cannot write to data output stream.");
			return null;
		}
	}

	//Prepares the body of an AK_SAY message, subject_id should be 10 digits long, detailed_msg_type should be 2 digits long,returns null if an exception is thrown
	
	final byte [] prepare_AK_SAY_MsgBody(int senderID,int detailed_msg_type, int subject)
	{
	
		try{
		int i;
	
		MessageParser m_msgParser=new MessageParser();
		
		String d_m_t=String.valueOf(detailed_msg_type);
		String s_sub=String.valueOf(subject);
		
			
		ByteArrayOutputStream baos= new ByteArrayOutputStream();
		DataOutputStream dos= new DataOutputStream(baos);
		
			
			dos.writeInt(senderID);         		//who sends the message
			dos.writeInt(msgtype_len+subj_len);						//Size of message MANDATORY for kernel in BYTES
			
			for (i=0;i<msgtype_len-d_m_t.length();i++)	//This portion needs to be STRING
				dos.writeByte(48);
			dos.writeBytes(d_m_t);
			for (i=0;i<subj_len-s_sub.length();i++)
				dos.writeByte(48);
			dos.writeBytes(s_sub);
			
			return baos.toByteArray();
			
		}catch(Exception e){
			System.out.println("IO prepare_AK_SAY_MsgBody: Cannot write to data output stream.");
			return null;
		}
		
	}
	
	//Prepares the body of an AK_MOVE message,returns null if an exception is thrown

	final byte [] prepare_AK_MOVE_MsgBody(int senderID,int [] route)
	{
		int i;
		ByteArrayOutputStream baos= new ByteArrayOutputStream();
		DataOutputStream dos= new DataOutputStream(baos);
		int route_max=ROUTER_SEND_MAX_LEVELS;
		if (route.length<route_max) route_max=route.length;
		
		try{
	
			dos.writeInt(senderID);         		//who sends the message
			for (i=0;i<route_max;i++)
				dos.writeInt(route[i]);				//ID of route element
			dos.writeInt(0)		;					//End of IDs
			return baos.toByteArray();

		}catch(Exception e){
			System.out.println("IO prepare_AK_MOVE_MsgBody: Cannot write to data output stream.");
			return null;
		}

	}

	//Prepares and sends AK_CONNECT message, returns false upon any failure
	final boolean sendAK_CONNECT(int agent_type)
	{
		byte [] msg_body;
		byte [] msg_packet;
		
		ByteArrayOutputStream baos= new ByteArrayOutputStream();
		DataOutputStream dos= new DataOutputStream(baos);
		try{

			dos.writeInt(0);		//Temporary ID
			dos.writeInt(0);		//Version
			dos.writeInt(agent_type);	
			msg_body=baos.toByteArray();	
			msg_packet=preparePacket(AK_CONNECT, msg_body );
			if (msg_packet!=null)
			{
				if (!sendPacket(msg_packet)) 
					return false;	
				else
					return true;
			}
			else 
			{
				return false;
			}
		}catch(Exception e){
			System.out.println("IO sendAK_CONNECT: Cannot write to data output stream/send AK_CONNECT.");
			return false;
		}	
	}

	final int receiveKA_CONNECT_OK(WorldModel agents_world)
	{
		int [] message_body;
		int message_body_byte_size;
		int message_type;
		int agent_id;
		MessageParser m_msgParser;

		m_msgParser=new MessageParser();
		message_body=receiveLongPacket();

		if (message_body!=null)
		{
			message_type=m_msgParser.get_msg_type_long_packet(message_body);
			if (message_type!= KA_CONNECT_OK)
			{
				System.out.println("receiveKA_CONNECT_OK:Could not receive connect ok message.");
				return -1;
			}
			agent_id=message_body[MSG_FORMAT_CONNECT_ID_INDEX];
			try{
			agents_world.parseKA_SENSE(message_body,(short)0);
			}catch(Exception e){}
			return agent_id;
		}
		else return -1;
	}

	final boolean sendAK_ACKNOWLEDGE(int agent_id)
	{
		byte [] msg_body;
		byte [] msg_packet;
		
		ByteArrayOutputStream baos= new ByteArrayOutputStream();
		DataOutputStream dos= new DataOutputStream(baos);

		try{	
			dos.writeInt(agent_id);	
			msg_body=baos.toByteArray();
			msg_packet=preparePacket(AK_ACKNOWLEDGE, msg_body );
			if (msg_packet!=null)
			{
				if (!sendPacket(msg_packet)) 
					return false;	
				else
					return true;
			}
			else 
			{
				return false;
			}			
		}catch(Exception e)
		{
			System.out.println("IO sendAK_ACKNOWLEDGE: Cannot write to data output stream/send AK_CONNECT.");
			return false;
		}
	}

	final boolean sendAK_EXTINGUISH(int agent_id, int target_id, int direction, int agent_x, int agent_y,int quantity)
	{
		byte [] msg_body;
		byte [] msg_packet;
		
		ByteArrayOutputStream baos= new ByteArrayOutputStream();
		DataOutputStream dos= new DataOutputStream(baos);

		try{	
			dos.writeInt(agent_id);	
			dos.writeInt(target_id);	
			dos.writeInt(direction);
			dos.writeInt(agent_x);	
			dos.writeInt(agent_y);		
			dos.writeInt(quantity);	
			dos.writeInt(0);	

			msg_body=baos.toByteArray();
			msg_packet=preparePacket(AK_EXTINGUISH, msg_body );
			if (msg_packet!=null)
			{
				if (!sendPacket(msg_packet)) 
					return false;	
				else
					return true;
			}
			else 
			{
				return false;
			}			
		}catch(Exception e)
		{
			System.out.println("IO sendAK_EXTINGUISH: Cannot write to data output stream/send AK_EXTINGUISH.");
			return false;
		}
	}	

	final boolean send_two_int_msg(int msg_type,int agent_id, int target_id)
	{
		byte [] msg_body;
		byte [] msg_packet;
		
		ByteArrayOutputStream baos= new ByteArrayOutputStream();
		DataOutputStream dos= new DataOutputStream(baos);

		try{	
			dos.writeInt(agent_id);	
			dos.writeInt(target_id);	
			
			msg_body=baos.toByteArray();
			msg_packet=preparePacket(msg_type, msg_body );
			if (msg_packet!=null)
			{
				if (!sendPacket(msg_packet)) 
					return false;	
				else
					return true;
			}
			else 
			{
				return false;
			}			
		}catch(Exception e)
		{
			System.out.println("IO send_two_int_msg: Cannot write to data output stream/send msg_type.");
			return false;
		}
	}

	final boolean send_one_int_msg(int msg_type,int agent_id)
	{
		byte [] msg_body;
		byte [] msg_packet;
		
		ByteArrayOutputStream baos= new ByteArrayOutputStream();
		DataOutputStream dos= new DataOutputStream(baos);

		try{	
			dos.writeInt(agent_id);		
			
			msg_body=baos.toByteArray();
			msg_packet=preparePacket(msg_type, msg_body );
			if (msg_packet!=null)
			{
				if (!sendPacket(msg_packet)) 
					return false;	
				else
					return true;
			}
			else 
			{
				return false;
			}			
		}catch(Exception e)
		{
			System.out.println("IO send_one_int_msg: Cannot write to data output stream/send msg_type.");
			return false;
		}
	}

	final void sendAK_SAY(int id, int detail_type, int subject)
	{
		
		byte [] msg_body;
		byte [] msg_packet;
		
		//send message
		msg_body=prepare_AK_SAY_MsgBody(id,detail_type, subject);
		if (msg_body!=null)
		{			
			msg_packet=preparePacket(AK_SAY, msg_body );
			if (msg_body!=null)
				sendPacket(msg_packet);
		}
		
	}
	
	final byte[] sendAK_TELL(int id, int detail_type, int subject)
	{
		
		byte [] msg_body;
		byte [] msg_packet;
		
		//send message
		msg_body=null;
		msg_packet=null;
		msg_body=prepare_AK_SAY_MsgBody(id,detail_type, subject);
		if (msg_body!=null)
		{
			msg_packet=preparePacket(AK_TELL, msg_body );
			if (msg_body!=null)
				sendPacket(msg_packet);
		}
		
		return msg_packet;
		
	}
	
	final byte[] sendAK_TELL(int id, int detail_type, int subject, int agent)
	{

		byte [] msg_body;
		byte [] msg_packet;

		//send message
		msg_body=null;
		msg_packet=null;
		msg_body=prepare_AK_SAY_MsgBody_Long(id,detail_type, subject, agent);
		if (msg_body!=null)
		{
			msg_packet=preparePacket(AK_TELL, msg_body );
			if (msg_body!=null)
				sendPacket(msg_packet);
		}

		return msg_packet;

	}
	
	final byte [] prepare_AK_SAY_MsgBody_Long(int senderID,int detailed_msg_type, int subject, int agent)
	{

		try{
		int i;

		MessageParser m_msgParser=new MessageParser();

		String d_m_t=String.valueOf(detailed_msg_type);
		String s_sub=String.valueOf(subject);
		String s_agent=String.valueOf(agent);

		ByteArrayOutputStream baos= new ByteArrayOutputStream();
		DataOutputStream dos= new DataOutputStream(baos);


			dos.writeInt(senderID);         		//who sends the message
			dos.writeInt(msgtype_len+subj_len + agent_len);						//Size of message MANDATORY for kernel in BYTES

			for (i=0;i<msgtype_len-d_m_t.length();i++)	//Right alignment yapmak iin
				dos.writeByte(48);
			dos.writeBytes(d_m_t);
			for (i=0;i<subj_len-s_sub.length();i++)
				dos.writeByte(48);
			dos.writeBytes(s_sub);
			for (i=0;i<agent_len-s_agent.length();i++)
				dos.writeByte(48);
			dos.writeBytes(s_agent);

			return baos.toByteArray();

		}catch(Exception e){
			System.out.println("IO prepare_AK_SAY_MsgBody: Cannot write to data output stream.");
			return null;
		}

	}

	final void sendAK_MOVE(int id, int [] route)
	{
		
		byte [] msg_body;
		byte [] msg_packet;
		
		//send message
		msg_body=prepare_AK_MOVE_MsgBody(id,route);
		if (msg_body!=null)
		{
			msg_packet=preparePacket(AK_MOVE, msg_body );
			if (msg_body!=null)
				sendPacket(msg_packet);
		}
		
	}


}


