import java.util.*;
import java.io.*;

class RouteFinder implements Definitions
{
	int m_current_level;           //Index indicating the current level of the tree
	Vector m_route_tree;           //Each element is a hashtable containing one level of the tree
	Vector m_position_history;     //Keeps the path passed while going to a target, when target changes this is reset

	Vector m_jammed_roads;         //roads that could not be reached even after sending block and int route
	Vector m_jammed_nodes;         //nodes that could not be reached even after sending int route
	Vector current_job_jam;        //a possible jam for a police officer
	Vector permanent_job_jam;      //a real jam for a police officer
	Vector m_encountered_blocked_roads; //The blocked roads hit during route search
	MessageParser m_msgParser;
	String filename;
	
	int prev_start;                //Route start id in previous call
	int prev_end;                  //Route end id in previous call
	int m_agent_index;	       //The type of the agent calling this
	boolean int_route_sent;      //Did we send a int route for a blocked road/node
	boolean block_sent;            //Did we send a block for a blocked road
	boolean consider_road_jamming;
	boolean consider_node_jamming;
	boolean fourth_condition;
	String reason;

        //The following are carried from FindRouteBetween to here in order to optimize
        	
	RouteResult routeReply;
	int idStart;
	SimObject endObject;
		
				
	Hashtable current_level_nodes;  	//The nodes in the current tree level (the ones that will be expanded)
	Hashtable next_level_nodes;		//The results of current expansions
	RouteNode current_node;		
	SimObject current_object;
	Integer temp_index;
	int parent_id;			        //The id of the current node being investigated
	int [] the_route;			//The found route
	int i;
	boolean found;                          //Did we find the target
	boolean expand;                         //Shall we expand this node
        boolean reset_next_cycle;               //Jam indicators reset for police officer if this is set in AgentThd or PoliceTeam (after sending clear for a road)
        boolean job_jam;
			
       	int available_lanes = 0;
        int total_lanes = 0;                 
                        
	RouteFinder()
	{
		m_route_tree=new Vector();		//Each element of vector consists of a hashtable
		m_position_history=new Vector();	//Elements visited by the agent, in reverse order. 0th element contains the last visited node.

		m_jammed_roads=new Vector();
		m_jammed_nodes=new Vector();
		current_job_jam=new Vector();
		permanent_job_jam=new Vector();
		m_msgParser=new MessageParser();
                routeReply=new RouteResult();
		
		m_encountered_blocked_roads=new Vector();
		int_route_sent=false;			
		block_sent=false;				
		prev_start=-1;
		prev_end=-1;
                reset_next_cycle=false;
                job_jam = false;
	}
	
	void setFilename(String x_filename, int agent_index)
	{
		filename=x_filename;
		m_agent_index=agent_index;
	}

	RouteResult FindRouteBetween(SimObject startObject,int idEnd, WorldModel the_world_model,boolean removelast, boolean withBlockedRoad, double PosExtra)
	//The start and end ids are static objects
	{
                
                m_route_tree.removeAllElements();	//We shall find a new route at each call
	        m_encountered_blocked_roads.clear();
                idStart=startObject.get_id();
                
        	found=false;
                expand=true;
                parent_id=-1;
                current_node=null;
		fourth_condition=false;     

		routeReply.routerResponse = NULL_ROUTE;
		routeReply.route = null;
		routeReply.blocked = -1; 
                routeReply.blockedNeighbor = -1;
                job_jam = false;
                
                if ((startObject instanceof Road) && permanent_job_jam.contains(new Integer(((Road)startObject).get_head()))  && permanent_job_jam.contains(new Integer(((Road)startObject).get_tail())))
                {
                        permanent_job_jam.remove(new Integer(((Road)startObject).get_head()));
                        permanent_job_jam.remove(new Integer(((Road)startObject).get_tail()));
                        System.out.println("############Police saved from self block....");
                }

                if (R_LOGGING && the_world_model.m_id == 2338) m_msgParser.writeMyFile(filename,"Routefinder called with soruce:" + idStart + "dest" + idEnd + "prevsource" + prev_start + "prevdest" + prev_end);
	
            
		//if (idStart==prev_start && idEnd==prev_end && !reset_next_cycle)
                //CASE 2: We could not move at all
                if (idStart==prev_start && !reset_next_cycle)
		{				
			
                  //CASE 2b:The police is close to a target, but it cannot reach
                  //This case occurs when the target is is a blocked road with a person on it, and for some reason our police cannot get near it
			//If size == 3 it means the target is 2 steps away
			if ( m_agent_index==WORLD_MODEL_POLICE_INDEX && m_position_history.size()==3)
			{ 
				//middle_route_node: The node that we cannot pass through
				Integer middle_route_node = new Integer(((RouteNode)m_position_history.elementAt(m_position_history.size()-2)).Id);	
                                Integer end_route_node = new Integer(((RouteNode)m_position_history.elementAt(m_position_history.size()-1)).Id);	
			
                                if (isthereaCivilianHere(middle_route_node.intValue(),the_world_model) || isthereaCivilianHere(end_route_node.intValue(),the_world_model) )
                                {
                                        //DROP YOUR TARGET
                                        //This if means: You have tried and could not pass the other side
                                        //Now you try this side and cannot pass too
				    if (current_job_jam.size()==1 && !(current_job_jam.contains(middle_route_node)))
				    {
					if (R_LOGGING && the_world_model.m_id == 2338) m_msgParser.writeMyFile(filename,"I will drop my target.");
					permanent_job_jam.add(current_job_jam.get(0));
					permanent_job_jam.add(middle_route_node);
					current_job_jam.removeAllElements();
					routeReply.routerResponse=DROP_JOB;
					routeReply.blocked=idEnd;
					return	routeReply;
				    }
                                        //Add the current edge you cannot pass through to the current_job_jam
				    else if (!current_job_jam.contains(middle_route_node))
				    {
					if (R_LOGGING && the_world_model.m_id == 2338) m_msgParser.writeMyFile(filename,"One step before dropping target.");
					current_job_jam.add(middle_route_node);
                                        job_jam = true;
				    }
                                }
                                else
                                {
                                        if (R_LOGGING && the_world_model.m_id == 2338) m_msgParser.writeMyFile(filename,"Job jam refused since no one around");
                                }
			}
                  //CASE 2c: We send block and int route for our immediate neighbors(This part assumes that for some reason we cannot se a block) 
			//This portion should work only if the above did not work
			
                        if (m_position_history.size()>1 && !job_jam)
			{
			//ILK NODU CIKARIRSAN BURASI DA DEGISECEK
			int first_element=((RouteNode)m_position_history.elementAt(m_position_history.size()-1)).Id;
			RouteNode second_element_node=(RouteNode)m_position_history.elementAt(m_position_history.size()-2);
			
			//If our immediate target is a road
			if (second_element_node.theNodeObject instanceof Road)
			{
				if (block_sent)
				{
					m_jammed_roads.add(new Integer(second_element_node.Id));
					if (R_LOGGING && the_world_model.m_id == 2338) m_msgParser.writeMyFile(filename,"FindRoute:Road jam considering open " + second_element_node.Id );
					consider_road_jamming=true;
					block_sent=false;			//we have come to the max level, we shall restart
					int_route_sent=false;
					
				}
				else if (int_route_sent) //We will send block
				{
					if (R_LOGGING && the_world_model.m_id == 2338) m_msgParser.writeMyFile(filename,"FindRoute:Road block send for " + second_element_node.Id);
					routeReply.routerResponse=CLEAR_ROUTE;
					routeReply.blocked=second_element_node.Id;
					block_sent=true;
					return routeReply; 
				}
				else
				{
					//Send int route
					//ILK NODU CIKARIRSAN BURASI DA DEGISECEK
					//If we send only 1road + 1node then it won't move
					if (m_position_history.size()>2)
					{
						int third_element=((RouteNode)m_position_history.elementAt(m_position_history.size()-3)).Id;		
						routeReply.route= new int[3];
						routeReply.route[0]=first_element;
						routeReply.route[1]=second_element_node.Id;
						routeReply.route[2]=third_element;
						if (R_LOGGING && the_world_model.m_id == 2338) m_msgParser.writeMyFile(filename,"FindRoute:Road int route of length 3 " + routeReply.route[0] + " " + routeReply.route[1] + " " + routeReply.route[2]);
						int_route_sent=true;
						return routeReply;
					}
					else
					{
						routeReply.route= new int[2];
						routeReply.route[0]=first_element;
						routeReply.route[1]=second_element_node.Id;
                                                if (R_LOGGING && the_world_model.m_id == 2338) m_msgParser.writeMyFile(filename,"FindRoute:Road int route of length 2 " + routeReply.route[0] + " " + routeReply.route[1]);
						int_route_sent=true;
						return routeReply;
					}
					
				}
			}
			//If our immediate target is a node
			else if(second_element_node.theNodeObject instanceof Node)
			{
				if (int_route_sent)
				{
					m_jammed_nodes.add(new Integer(second_element_node.Id));		
					consider_node_jamming=true;
					if (R_LOGGING && the_world_model.m_id == 2338) m_msgParser.writeMyFile(filename,"FindRoute: Node jam considering open for " + second_element_node.Id);					
					int_route_sent=false;//we have come to the max level, we shall restart
				}
				else
				{
					//Send int route
					//Short route of length >2 //If we send only 1node + 1road then it won't move
					if (m_position_history.size()>2)
					{	
						int third_element=((RouteNode)m_position_history.elementAt(m_position_history.size()-3)).Id;								
						routeReply.route= new int[3];
						routeReply.route[0]=first_element;
						routeReply.route[1]=second_element_node.Id;
						routeReply.route[2]=third_element;
						if (R_LOGGING && the_world_model.m_id == 2338) m_msgParser.writeMyFile(filename,"FindRoute:we will send int route of length 3 for node " + routeReply.route[0] + " " + routeReply.route[1] + " " + routeReply.route[2]);
						int_route_sent=true;
						return routeReply;
					}
					//Short route of length 2
					else
					{
						routeReply.route= new int[2];
						routeReply.route[0]=first_element;
						routeReply.route[1]=second_element_node.Id;
                                                if (R_LOGGING && the_world_model.m_id == 2338) m_msgParser.writeMyFile(filename,"FindRoute:we will send int route of length 2 for node " + routeReply.route[0] + " " + routeReply.route[1]);
						int_route_sent=true;
						return routeReply;
					}//end of history length check

				}//End of int route sent check
			
			}//End of immediate target check
			
		      }//END OF CASE 2c (which is jamming checks)
		}//END OF CASE2 (which is no movement at all)
                //changed 21/11/2002 we do not reset jammed roads and nodes if we have not yet reached the destination
                //the following elseif was introduced
                else if (idEnd==prev_end && !reset_next_cycle)
                {
                        if (R_LOGGING && the_world_model.m_id == 2338) m_msgParser.writeMyFile(filename,"FindRoute: Reseting indices to be able to add new jams");
                        int_route_sent=false;
			block_sent=false;
                }
		else //If we have moved or we ignore that we could not move (resetnextcycle is true)
		{
                        if (R_LOGGING && the_world_model.m_id == 2338) m_msgParser.writeMyFile(filename,"FindRoute: We have moved or reseting removing all jams Reset: " + reset_next_cycle);
			int_route_sent=false;
			block_sent=false;
			m_jammed_roads.removeAllElements();
			m_jammed_nodes.removeAllElements();
                        consider_road_jamming=false;
		        consider_node_jamming=false;
		}
		
		
		//For police officers, job jam should continue until both initial position and target have changed
		if (idStart!=prev_start && idEnd!=prev_end)
		{
			if (R_LOGGING && the_world_model.m_id == 2338) m_msgParser.writeMyFile(filename,"Police changed job, dropping current job jams");
			current_job_jam.removeAllElements();
		}

		prev_start=idStart;
		prev_end=idEnd;

		//Add the root
		next_level_nodes=new Hashtable();
		next_level_nodes=AddChild(next_level_nodes, null, idStart);
		m_route_tree.add(next_level_nodes); 
		m_current_level=0;
	
//PATH FINDER START~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~		
//Repeat until we find a route or reach max path length(path length is limited by own decision, not by simulator)
//The levels go from 0..ROUTER_PREPARE_MAX_LEVELS
//ROUTER_PREPARE_MAX_LEVELS level is left unexpanded(unchecked) at the end of the loop
		while (m_current_level<ROUTER_PREPARE_MAX_LEVELS)		
            //Looping the levels of the tree
		{
			current_level_nodes=(Hashtable)m_route_tree.get(m_current_level); 	//Get the nodes at current level
			next_level_nodes=new Hashtable();						//Prepare the nodes for next leve
			Iterator it=current_level_nodes.keySet().iterator();	//Expand each node at this level
			while(it.hasNext())
			//Looping the nodes at current level for expanding
                        {
				expand=true;
			
				temp_index=(Integer)it.next();
				current_node=(RouteNode)current_level_nodes.get(temp_index);
				parent_id=current_node.Id;
				if (parent_id==idEnd)				//We have found the route
				{	
					found=true;
					break;
				}

				//Assigning the object of a routenode when it is about to be expanded
				current_object=the_world_model.getObjectbyId(parent_id);
				current_node.theNodeObject=current_object;		
				
				switch(current_object.get_type())
				{
				case TYPE_NODE: 
					//Expand each node immediately, necessary checks are done when added edges will be expanded
					int [] edges;	
                                        
                                        reason = "";
					edges =	((Node)current_object).get_edges();
						
					if ( !(current_job_jam.contains(new Integer(parent_id))) && !(permanent_job_jam.contains(new Integer(parent_id)))  && (!(m_jammed_nodes.contains(new Integer(parent_id))) || !consider_node_jamming ))
					{
                                           //We do not add this node to the tree if it is an immediate neigbor of the source and there's someone there
                                           
                                     
					   if (m_current_level==1 )
					   {
					   
					        // if (R_LOGGING) m_msgParser.writeMyFile(filename,"This node is expandable initially");
						if (isthereaCivilianHere(parent_id,the_world_model))
						{
							//System.out.println("Komsu nodda biri var gitmiyoruz Nod: " + parent_id+ " m_id:" + the_world_model.m_id);		
							expand=false;
                                                        reason = "Komsuda biri var";
						}
                                                
						if (startObject instanceof Road)
						{
							if (!the_world_model.clear(startObject))
							{
								//Routedaki ikinci elemana bakiyoruz, roadin headi mi taili mi diye...
								int head_tail_node=parent_id;	
								int other_end_node=0;
								SimObject other_end_object;
						
								//if (R_LOGGING) m_msgParser.writeMyFile(filename,"Head-tail node" + head_tail_node);
								//if (R_LOGGING) m_msgParser.writeMyFile(filename,"startin headi " + ((Road)startObject).get_head());
								//if (R_LOGGING) m_msgParser.writeMyFile(filename,"startin taili "+ ((Road)startObject).get_tail());
						
								if (head_tail_node==((Road)startObject).get_head())
								{
									other_end_node=((Road)startObject).get_tail();
								}
								else if(head_tail_node==((Road)startObject).get_tail())
								{
									other_end_node=((Road)startObject).get_head();
								}
						
								//if (R_LOGGING) m_msgParser.writeMyFile(filename,"other end node "+ other_end_node);
								other_end_object=(SimObject)the_world_model.getObjectbyId(other_end_node);
						
								if (other_end_object == null)
								{
									System.out.println(filename + "Road in head ya da tail objectini bulamadik " + other_end_node + " road is" + startObject.m_id + "head_tail_node is" + head_tail_node);
								}
				                               
								if (!(the_world_model.passable(startObject,other_end_object, PosExtra)))
								{
                                                                        if ( !withBlockedRoad)
                                                                        {                        
									       //System.out.println("Yehoo baslangic road to node non-passable oldugu icin expand edilmedi..." + startObject.get_id());
									       //The first element of the route is reported if it is not passable in the direction we wanted to move in the previous cycle
									       if (m_position_history.contains(current_node))
									       {
									         	routeReply.routerResponse=CLEAR_ROUTE;
									        	routeReply.blocked=startObject.m_id;
									       }								
									       expand=false;
                                                                               reason = "Ilk node not passable";
                                                                        }
                                                                        else //THIS IS A POLICE OFFICER
                                                                        {
                                                                                routeReply.blockedNeighbor=startObject.m_id;
                                                                        }
								}
							}//End of start object clear check
						}//End of start object road check	                        
					   }
					   if (expand)
					   {
						for (i=0;i<edges.length;i++)
						{
						         if (notEncountered(edges[i],m_current_level))
							 {
							        //if (R_LOGGING) m_msgParser.writeMyFile(filename,"EXPANDING NODE " +parent_id + " edge" +edges[i]);
								next_level_nodes=AddChild(next_level_nodes, current_node, edges[i]);
							 }
						}
					   }
					   else
					   {
					           if (R_LOGGING && the_world_model.m_id == 2338) m_msgParser.writeMyFile(filename,"Expand false for node: " + parent_id+ " m_id:" + the_world_model.m_id + "reason: " + reason);   
					   }
				    }//End of first expandability check...
				    else
				    {  
                                        if (R_LOGGING && the_world_model.m_id == 2338) m_msgParser.writeMyFile(filename,"The node is: " + parent_id+ "current_job_jam.contains: " + current_job_jam.contains(new Integer(parent_id)) + "permanent_job_jam.contains: " + permanent_job_jam.contains(new Integer(parent_id)) + "m_jammed_nodes.contains: " + m_jammed_nodes.contains(new Integer(parent_id)) + "consider_node_jamming: " + consider_node_jamming);
                                    }   				    
				break;
				case TYPE_BUILDING:case TYPE_POLICE_OFFICE : case TYPE_FIRE_STATION: case TYPE_REFUGE:
				case TYPE_AMBULANCE_CENTER:

					//Do not expand a burning building, i.e. do not add it to route--26/10/2002This is taken back:
                                        //Otherwise,if someone enters a burning building it cannot go out!!!
					//Expanding a building only if it is the origin 19/04/2002 -- traf sim 
					if ( m_current_level==0)
					{
						int [] entrances;
						entrances=((Building)current_object).get_entrances();
						for (i=0;i<entrances.length;i++)
						{
							if (notEncountered(entrances[i],m_current_level))
								next_level_nodes=AddChild(next_level_nodes, current_node, entrances[i]);
						}
					}
					else
					{
						//if (R_LOGGING) m_msgParser.writeMyFile(filename,"Burning building rejected " + current_object.get_id());
					}
				break;
				case TYPE_ROAD:
					//Don not expand a non-passable road
					//find a route that does not contain a blocked road
					
                                        //If there's someone on the road and this road is an immediate neighbor of the source
					reason = "";
					if (m_current_level==1)
					{
						//If route's second element is non -passable and we tried to get there previously, we report it
						if (!(the_world_model.passable(current_object,current_node.Parent.theNodeObject)))
						{			
							if (m_position_history.contains(current_node))
							{
								routeReply.routerResponse=CLEAR_ROUTE;
								routeReply.blocked=parent_id;
							}
                                                        
                                                        if (!withBlockedRoad) 
                                                                expand = false;	                                                                                          
                                                        else
                                                                routeReply.blockedNeighbor = parent_id;
                                                                
                                                        reason = "Level 1 Not passable";
						}
					        else
						{/*AGENTS CAN MOVE FROM NODES TO NEIGHBOR ROADS 1LANE<>1PERSON, THEY CAN EXIST IN SAME LANE AND IN DIFFERENT DISTANCES
							available_lanes = 0;
                                                        total_lanes = 0;
                                                        
                                                        
							if (idStart==((Road)current_object).get_tail())
                                                        {
							         available_lanes = ((Road)current_object).get_aliveLinesToHead();
                                                                 total_lanes = ((Road)current_object).get_linesToHead();
							}
                                                        else if (idStart==((Road)current_object).get_head())	
							{
                                                                 available_lanes = ((Road)current_object).get_aliveLinesToTail();     
                                                                 total_lanes = ((Road)current_object).get_linesToTail();
                                                        }      
         
                                                        expand = !isthereaCivilianHere(parent_id, idStart, the_world_model, available_lanes);
                
                                                        //If we are a police and there are lanes that can be cleared then we should  clear those lanes
                                                        if (withBlockedRoad && !expand && total_lanes>available_lanes) 
                                                        {
                                                                routeReply.blockedNeighbor = parent_id;
                                                                expand=true;
                                                        }        
                
							if (!expand)
							{
								if (R_LOGGING) m_msgParser.writeMyFile(filename,"Komsuda roadda biri var oraya gitmiycez 1 " + parent_id);
								//System.out.println("Komsu roadda biri var gitmiyoruz Road: " + parent_id+ " m_id:" + the_world_model.m_id);		
								reason = "Komsuda biri var. Total lanes: " + total_lanes + "available_lanes: " + available_lanes;
							}		
			     	                       */
                                                }
					}
									
					if (expand && !withBlockedRoad) //for ambulance and firefighter
					{
						if(current_node.Parent != null)
						{
							expand=the_world_model.passable(current_object,current_node.Parent.theNodeObject);
							
							if (!expand)
							{
								//if (R_LOGGING) m_msgParser.writeMyFile(filename,"Road blocked oraya gitmiycez 1 " + parent_id);
								current_object.m_routelevel=m_current_level;
								m_encountered_blocked_roads.add(new BlockedRoad(m_current_level,current_object));
								reason = "Blocked";
							}	
						}
						
					}
					
					//we think we can go there but if there is a jam
					if (expand && consider_road_jamming)
					{
						expand=!(m_jammed_roads.contains(new Integer(parent_id)));
						if (!expand)		
        						reason = "Jamming";					
					}
					
					if (expand)
					{
						int tmp_id;
	
						tmp_id=((Road)current_object).get_head();
						if (notEncountered(tmp_id,m_current_level))
							next_level_nodes=AddChild(next_level_nodes, current_node,tmp_id );

						tmp_id=((Road)current_object).get_tail();
						if (notEncountered(tmp_id,m_current_level))
							next_level_nodes=AddChild(next_level_nodes, current_node,tmp_id);
					}
					else
					{
                                                if (R_LOGGING && the_world_model.m_id == 2338) m_msgParser.writeMyFile(filename,"Expand false for road: " + parent_id + "reason: " + reason);
					}
					
				break;
				default:
				    System.out.println("Router could not get object type....Sorry.");
				break; 
				}//Switch

			}//End of looping the nodes of the current level

			if (found) break;
			if (next_level_nodes.size()==0) break;	//We have no nodes to expand
			m_route_tree.add(next_level_nodes);
			m_current_level++;

		}//End of looping the levels of the current tree
//PATH FINDER END ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

           
	        //POSTCASE 1:We have found the target
        	//If we have found the route, we will form an integer array of ids.
                //Current level represents the level that is being expanded
                //If the target is found, it is found in the `current_level`
		if (found)
		{
                        if (R_LOGGING && the_world_model.m_id == 2338) m_msgParser.writeMyFile(filename,"**route found");
                        
			if (removelast)
			{
				routeReply.route=convertTreetoPath(m_current_level-1, current_node.Parent,withBlockedRoad);
			}
			else
			{
				routeReply.route=convertTreetoPath(m_current_level, current_node,withBlockedRoad);
			}
		}
            //POSTCASE 2:Max route length reached, no target found
		//If we could not find the route because of route length limitations
		//We will return a route to the reached endpoint that is closest to the endpoint we are aiming for
		else if (m_current_level==ROUTER_PREPARE_MAX_LEVELS)
		{
                        if (R_LOGGING && the_world_model.m_id == 2338) m_msgParser.writeMyFile(filename,"Route not found, max levels hit.");
			//Since we have not yet expanded the nodes in the current level, we shall look at one level above 24/4/2002
			//in this case m_current_level cannot be negative, no need to check
			current_level_nodes=(Hashtable)m_route_tree.get(m_current_level-1);
			endObject=the_world_model.getObjectbyId(idEnd);
			current_node=the_world_model.findClosestRouteNode(current_level_nodes,endObject);
			//System.out.println(filename + "3");
			routeReply.route=convertTreetoPath(m_current_level-1,current_node,withBlockedRoad);
		}
                
            //POSTCASE 3: We could not find the route and could not reach the max route length
            //In this case one of the blocked roads that is closest to the target is aimed                
		else
		{
                        if (R_LOGGING && the_world_model.m_id == 2338) m_msgParser.writeMyFile(filename,"Route not found, max levels not hit.");
			SimObject endObjtoSend;
			if (m_encountered_blocked_roads.size()>0)
			{
                                if (R_LOGGING && the_world_model.m_id == 2338) m_msgParser.writeMyFile(filename,"There are blocked ends to choose from.");
				endObject=the_world_model.getObjectbyId(idEnd);
				endObjtoSend=the_world_model.closestForRandomRoute(m_encountered_blocked_roads,endObject);
			
				if (endObjtoSend!=null)
				{	                                  
                                        routeReply.route = convertTreetoPath(endObjtoSend.m_routelevel, endObjtoSend.m_id,withBlockedRoad);
					if (routeReply.route.length<2 )
					{
						if (R_LOGGING && the_world_model.m_id == 2338) m_msgParser.writeMyFile(filename,"Not appropriate blocked road");
						routeReply.route = null;
						if (routeReply.blocked<0)
						{
							routeReply.routerResponse=CLEAR_ROUTE;
							routeReply.blocked=endObjtoSend.m_id;
                                                       
						}
						return routeReply;
					}
					else
					{
						if (R_LOGGING && the_world_model.m_id == 2338) m_msgParser.writeMyFile(filename,"Blocked road found" + endObjtoSend.m_id);
						if (routeReply.blocked<0)
						{
							routeReply.routerResponse=CLEAR_ROUTE;
							routeReply.blocked=endObjtoSend.m_id;
                                                       
						}
						return routeReply;
					}				
				}
				else
				{
                                        if (R_LOGGING && the_world_model.m_id == 2338) m_msgParser.writeMyFile(filename,"Appropriate blocked road not found 4th condition");
					fourth_condition = true;
				}
                        }
			else
			{
                                if (R_LOGGING && the_world_model.m_id == 2338) m_msgParser.writeMyFile(filename,"No blocked road 4th condition");
				fourth_condition = true;
			}

			//POSTCASE 4:
			//if we could not find a route in previous check, then current level contains non expandable nodes(not roads), we find a node at
			//the level before the current level(the non expandable nodes in this level may be burning buildings....)
			if (fourth_condition)
			{
				if (m_current_level>1)
				{
					RouteNode routeEnd;	
					routeEnd=getanExpandableParent((Hashtable)m_route_tree.get(m_current_level));

					if (routeEnd !=null)
					{	//System.out.println(filename + "5");
						routeReply.route = convertTreetoPath(m_current_level-1, routeEnd,withBlockedRoad);
						if (R_LOGGING && the_world_model.m_id == 2338) m_msgParser.writeMyFile(filename,"4th condition an expandaHle parent " + routeEnd.Id);
						m_jammed_roads.removeAllElements();
						m_jammed_nodes.removeAllElements();
                                                current_job_jam.removeAllElements();
					}
					else{
						m_jammed_roads.removeAllElements();
						m_jammed_nodes.removeAllElements();
                                                current_job_jam.removeAllElements();
						m_position_history.removeAllElements();
						if (R_LOGGING && the_world_model.m_id == 2338) m_msgParser.writeMyFile(filename,"4th condition no expandable parent NO ROUTE" );
						routeReply.routerResponse=NULL_ROUTE1;
						routeReply.blocked=-1;
						return routeReply;
					}
				}
				else
				{
					m_jammed_roads.removeAllElements();
					m_jammed_nodes.removeAllElements();
                                        current_job_jam.removeAllElements();
					m_position_history.removeAllElements();
                                        if (R_LOGGING && the_world_model.m_id == 2338) m_msgParser.writeMyFile(filename,"4th condition level <=1 NO ROUTE" );
                                        routeReply.routerResponse=NULL_ROUTE2;
					routeReply.blocked=-1;
					return routeReply;
				}
			}//END OF POSTCASE 4	
		}//END OF POSTCASE 3
		
		return routeReply;
	}


	Hashtable AddChild( Hashtable the_nodes, RouteNode the_parent, int own_id)
	{
		RouteNode current_node;		

		current_node= new RouteNode();
		current_node.Parent=the_parent;			
		current_node.Id=own_id;
		the_nodes.put(new Integer(own_id),current_node); 

		return the_nodes;
	}


	boolean notEncountered(int Id, int deepest_level)
	{
		int i;
		Integer search_id=new Integer(Id);
		Hashtable current_hashtable
;		
		//If during search we have expanded this node, return false
		for (i=deepest_level;i>=0;i--)
		{
			current_hashtable=(Hashtable)m_route_tree.get(i);
			if (current_hashtable.containsKey(search_id))
				{
					return false;
				}
		}
		
		return true;
	}
	
	int [] convertTreetoPath(int deepest_level,int current_id,boolean withBlockedRoad)
	{
	
		RouteNode current_node;
		Hashtable current_hashtable;
		
		current_hashtable=(Hashtable)m_route_tree.get(deepest_level);
		current_node=(RouteNode)current_hashtable.get(new Integer(current_id));
		
		return convertTreetoPath(deepest_level,current_node,withBlockedRoad);
		
	}
	

	int [] convertTreetoPath(int deepest_level,RouteNode current_node,boolean withBlockedRoad )
//Deepest_level is the upper bound of the tree level array, current_node is the last node
	{
		
		int i;
		int [] result;
		RouteNode temp_node=current_node;
		int route_level;
		
		m_position_history.removeAllElements();		
		route_level=deepest_level;

		//THIS DOES NOT REMOVE THE FIRST ELEMENT OF EACH ROUTE
		result=new int[deepest_level+1];	//Size of array is 1+upperbound
	
		for (i=route_level;i>=0;i--)
		{
			result[i]=current_node.Id;					//Write the id of the current node
			//if (R_LOGGING) m_msgParser.writeMyFile(filename,"step" + i + " " + result[i]);	
			//System.out.println("step" + i + " " + result[i]);
			m_position_history.add(current_node);		//Add the node to the position history
			
			current_node=(RouteNode)current_node.Parent;
		}
		
		return result;
	}
	
	
	void dumpRouteTree()
	{
	
		int i;
		Integer temp_index;
		Hashtable current_hashtable;
		RouteNode current_node;
		
		for (i=0;i<m_route_tree.size();i++)
		{		
			if (R_LOGGING) m_msgParser.writeMyFile(filename,"RouteFinder:dumpRouteTree:RouteDump level "+ i);
			current_hashtable=(Hashtable)m_route_tree.get(i);
			Iterator it=current_hashtable.keySet().iterator();	//Expand each node at this level
			
			while(it.hasNext())
			{
				temp_index=(Integer)it.next();
				current_node=(RouteNode)current_hashtable.get(temp_index);
				
				if (R_LOGGING) 
				{
					if (current_node.Parent != null)
					{
						//m_msgParser.writeMyFile(filename,"parent "+ current_node.Parent.Id + "this "+ current_node.Id); 		
						System.out.println("parent "+ current_node.Parent.Id + "this "+ current_node.Id);
					}
					else
					{
						//m_msgParser.writeMyFile(filename,"parent -"+  "this "+ current_node.Id);
						System.out.println("parent -"+  "this "+ current_node.Id); 		
					}
				}
			}
		}
	}
	
	//Scans the parents of the current level of the route tree, returns the first non-jammed parent
	RouteNode getanExpandableParent(Hashtable hash)
	{
	
		Integer child_index;
		Integer parent_index;
		RouteNode node;
		
		Iterator it=hash.keySet().iterator();
		while(it.hasNext())
		{
			child_index=(Integer)it.next();
			node=(RouteNode)hash.get(child_index);
			if (node.Parent != null)
			{
				parent_index=new Integer(node.Parent.Id);
				if (!(m_jammed_roads.contains(parent_index)))
					return (node.Parent);
			}
		}
		return null;
	}
	

	boolean isthereaCivilianHere(int position, int Startposition, WorldModel the_world_model, int aliveLines)
	{
		Integer toCount;
		
		toCount = (Integer)the_world_model.positionsofMovingObjs.get(new Integer (position));
		if (toCount ==null)
			toCount = new Integer(0);
	
		if (toCount.intValue()+ 1 > aliveLines)
                {
                        System.out.println("To: " + position + "count: " + toCount.intValue() + the_world_model.m_id );
			return true;            
		}
                else if (the_world_model.positionsofMovingObjs.containsKey(new Integer (position)))
			System.out.println("****************************Otherwise non passable road becomes passable" + the_world_model.m_id);
		return false;
	}
	
	boolean isthereaCivilianHere(int position, WorldModel the_world_model)
	{
		return the_world_model.positionsofMovingObjs.containsKey(new Integer (position));
	}
	
}



class RouteNode
{

	int Id;			//ID of this node
	RouteNode Parent;	//'the' parent node
	SimObject theNodeObject; //Reference to the object itself

}

class RouteResult implements Definitions
{
	int routerResponse;
	int [] route;
	int blocked;
        int blockedNeighbor;

	RouteResult()
	{
		routerResponse=NULL_ROUTE;
		blocked=-1;
		route= null;
                blockedNeighbor=-1;
	}
}

class BlockedRoad
{
        int level;
        SimObject blocked_road;
        
        BlockedRoad(int x_level, SimObject x_blocked_road)
        {
               level= x_level;
               blocked_road=x_blocked_road;
        
        }

}

